from hgvs.pretty.models import VariantCoords, VariantData
from hgvs.pretty.console.renderer import BasicRenderer
from hgvs.sequencevariant import SequenceVariant


class RegionImpacted(BasicRenderer):
    """
    A class used to represent the region impacted by a variant in a genomic sequence.

    Attributes
    ----------
    var_g : SequenceVariant
        The genomic variant to be displayed.
    vc : VariantCoords
        The coordinates of the variant.
    Methods
    -------
    legend() -> str
        Returns a string representing the legend for the variant display.
    display(data: VariantData) -> str
        Returns a string representing the shuffled variant display based on the provided data.
    """

    def __init__(
        self, config, orientation: int, var_g: SequenceVariant, vc: VariantCoords
    ) -> None:
        super().__init__(config, orientation)

        self.var_g = var_g
        self.vc = vc

    def legend(self) -> str:
        return "region    : "

    def display(self, data: VariantData) -> str:
        from hgvs.pretty.console.constants import ENDC, COLOR_MAP

        seq_start = data.display_start
        seq_end = data.display_end

        split_char = "|"

        start = self.vc.start
        end = self.vc.end

        # map interbase coordinates to 1-based display coords:
        start = start + 1

        if self.var_g.posedit.edit.type == "sub":
            end = start
            if len(self.vc.alt) == 1:
                split_char = self.vc.alt
        vlen = end - start + 1
        if self.var_g.posedit.edit.type == "ins" and vlen == 0:
            start = start - 1
            end = end + 1
            split_char = "^"

        if self.var_g.posedit.edit.type == "del":
            split_char = ""
            if self.config.use_color:
                split_char = COLOR_MAP["del"]
            split_char += "x"
            if self.config.use_color:
                split_char += ENDC

        elif self.var_g.posedit.edit.type == "identity":
            split_char = "="

        if start < seq_start:
            # raise ValueError(f"Can't create shuffled representation, since start {start} < seq_start {seq_start} ")
            return ""
        if end > seq_end:
            return ""
            # raise ValueError(f"Can't create shuffled representation, since end {end} > seq_end {seq_end} ")

        var_str = ""
        in_range = False

        reverse_display = False
        if self.orientation < 0 and self.config.reverse_display:
            reverse_display = True

        for pdata in data.position_details:
            p = pdata.chromosome_pos
            if not p:
                if in_range:
                    var_str += "-"
                else:
                    var_str += " "
                continue

            if not reverse_display and p == start:
                var_str += split_char
                in_range = True
            elif reverse_display and p == end:
                var_str += split_char
                in_range = True
            elif not reverse_display and p == end:
                var_str += split_char
                in_range = False
            elif reverse_display and p == start:
                var_str += split_char
                in_range = False
            elif not reverse_display and p > end and in_range:
                in_range = False
                var_str += " "
            elif reverse_display and p < start and in_range:
                in_range = False
                var_str += " "
            elif in_range:
                var_str += "-"
            else:
                var_str += " "

        return var_str



"""simplified imports for the hgvs package

`hgvs.easy` simplifies using the hgvs package by providing a single
import path and objects that are instantiated with common defaults.

With default logging levels, using hgvs.easy is straightforward and
requires no code changes::

    >> from hgvs.easy import parser, projector
    >> var_g = parser.parse("NC_000017.11:g.43091687delC")
    >> projector.relevant_transcripts(var_g)
    ['NM_007294.3', 'NM_007297.3', 'NR_027676.1', 'NM_007298.3', 'NM_007299.3', 'NM_007300.3']
    >> projector.g_to_t(var_g, "NM_007294.3")
    SequenceVariant(ac=NM_007294.3, type=c, posedit=3844del)

`hgvs.easy` also introduces new functional forms for common methods.
For example::

    >> from hgvs.easy import parse, get_relevant_transcripts, g_to_t
    >> var_g = parse("NC_000017.11:g.43091687delC")
    >> get_relevant_transcripts(var_g)
    ['NM_007294.3', 'NM_007297.3', 'NR_027676.1', 'NM_007298.3', 'NM_007299.3', 'NM_007300.3']
    >> g_to_t(var_g, "NM_007294.3")
    SequenceVariant(ac=NM_007294.3, type=c, posedit=3844del)


NOTE: A consequence of making imports easy is a loss of
configurability by the caller.  The database connection is made with
no arguments (i.e., `connect()`), so it honors the UTA_DB_URL and
HGVS_SEQREPO_DIR environment variables but is otherwise not
configurable by the caller.

"""

from hgvs import __version__, global_config  # noqa: F401
from hgvs.assemblymapper import AssemblyMapper
from hgvs.dataproviders.uta import connect
from hgvs.normalizer import Normalizer
from hgvs.parser import Parser
from hgvs.pretty.prettyprint import PrettyPrint
from hgvs.validator import Validator
from hgvs.variantmapper import VariantMapper

# provide standard abbreviated, short, and long names for instances
hp = parser = hgvs_parser = Parser()
hdp = hgvs_data_provider = connect()
vm = variant_mapper = hgvs_variant_mapper = VariantMapper(hgvs_data_provider)
am37 = hgvs_assembly_mapper_37 = AssemblyMapper(
    hgvs_data_provider, assembly_name="GRCh37"
)
am38 = projector = hgvs_assembly_mapper_38 = AssemblyMapper(
    hgvs_data_provider, assembly_name="GRCh38"
)
hn = normalizer = hgvs_normalizer = Normalizer(hgvs_data_provider)
hv = validator = hgvs_validator = Validator(hgvs_data_provider)

# functionalized forms of common methods
parse = parser.parse
normalize = normalizer.normalize
validate = validator.validate
c_to_g = projector.c_to_g
c_to_n = projector.c_to_n
c_to_p = projector.c_to_p
g_to_c = projector.g_to_c
g_to_n = projector.g_to_n
g_to_t = projector.g_to_t
n_to_c = projector.n_to_c
n_to_g = projector.n_to_g
t_to_g = projector.t_to_g
t_to_p = projector.t_to_p
get_relevant_transcripts = am38.relevant_transcripts
pretty37 = PrettyPrint(hdp, am37, use_color=True, show_legend=True)
pretty38 = PrettyPrint(hdp, am38, use_color=True, show_legend=True)

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from dataclasses import dataclass
from typing import List

from hgvs.alignmentmapper import AlignmentMapper
from hgvs.assemblymapper import AssemblyMapper
from hgvs.location import Interval
from hgvs.sequencevariant import SequenceVariant
from hgvs.dataproviders.interface import Interface


@dataclass(eq=True, repr=True, frozen=True, order=True)
class VariantCoords:
    """Representation of a variant in one of the shuffled representation"""

    start: int
    end: int
    ref: str
    alt: str


@dataclass(eq=True, frozen=True, order=True)
class ProteinData:
    c_pos: int = None
    aa: str = None  # one letter AA code
    tlc: str = None  # three letter AA code
    aa_char: str = None  # the character from the tlc that maps to the c_pos
    var_p: SequenceVariant = None
    is_init_met: bool = False
    is_stop_codon: bool = False
    aa_pos: int = -1

    @classmethod
    def get_header(cls) -> str:
        return "c_pos\ttlc\taa_char\tc_pos\taa_pos"

    def __repr__(self) -> str:
        return (
            f"{self.c_pos}\t{self.tlc}\t{self.aa_char}\t{self.c_pos % 3}\t{self.aa_pos}"
        )


@dataclass(eq=True, frozen=False, order=True)
class PositionDetail:
    chromosome_pos: int = None
    alignment_pos: int = None
    mapped_pos: int = None
    mapped_pos_offset: int = None
    cigar_ref: str = None
    n_pos: int = None
    c_pos: int = None
    c_offset: int = None
    ref: str = None
    tx: str = None
    n_interval: Interval = None
    c_interval: Interval = None
    protein_data: ProteinData = None
    exon_nr: int = None
    variant_feature: str = None

    @classmethod
    def get_header(cls) -> str:
        return (
            "alignment_pos\tc_pos\tc_offset\tc_interval\tcigar_ref\tchromosome_pos\tref\tmapped_pos\t"
            "mapped_pos_offset\tn_pos\ttx\tvariant_feature\texon_nr"
        )

    def __repr__(self) -> str:
        return (
            f"{self.alignment_pos}\t{self.c_pos}\t{self.c_offset}\t{self.c_interval}\t"
            + f"{self.cigar_ref}\t"
            + f"{self.chromosome_pos}\t{self.ref}\t{self.mapped_pos}\t{self.mapped_pos_offset}\t"
            + f"\t{self.n_pos}\t{self.tx}"
            + f"\t{self.variant_feature} {self.exon_nr}"
        )


@dataclass(eq=True, repr=True, frozen=True, order=True)
class VariantData:
    """
    A data container for knowledge we have about a variant.

    Attributes:
        display_start (int): The start position of the variant for display purposes.
        display_end (int): The end position of the variant for display purposes.
        left_shuffled (VariantCoords): The coordinates of the variant after left shuffling.
        right_shuffled (VariantCoords): The coordinates of the variant after right shuffling.
        fully_justified (VariantCoords): The coordinates of the variant after full justification.
        display_seq (str): The sequence of the variant for display purposes.
        tx_seq (str): The transcript sequence of the variant.
        alignmentmapper (AlignmentMapper): The alignment mapper associated with the variant.
        var_g (SequenceVariant): The genomic sequence HGVS representation of the variant.
        strand (int): The strand of the variant (1 for positive, -1 for negative).
        var_c_or_n (SequenceVariant, optional): The coding or non-coding sequence HGVS representation of the variant. Defaults to None.
        var_p (SequenceVariant, optional): The protein sequence HGVS representation of the variant. Defaults to None.
        position_details (List[PositionDetail], optional): The list of position details associated with the variant. Defaults to None.
        all (bool, optional): A flag indicating if all overlapping transcripts should be included. Defaults to False.
        is_rna (bool, optional): A flag indicating if the variant is RNA. Defaults to False.
    """

    display_start: int
    display_end: int
    left_shuffled: VariantCoords
    right_shuffled: VariantCoords
    fully_justified: VariantCoords
    display_seq: str
    tx_seq: str
    alignmentmapper: AlignmentMapper
    var_g: SequenceVariant
    strand: int
    var_c_or_n: SequenceVariant = None
    var_p: SequenceVariant = None
    position_details: List[PositionDetail] = None
    all: bool = False
    is_rna: bool = False


@dataclass
class PrettyConfig:
    """A container for various configurations."""

    hdp: Interface
    assembly_mapper: AssemblyMapper
    padding_left: int = 20
    padding_right: int = 20
    use_color: bool = False
    show_legend: bool = True
    show_all_shuffleable_regions = False
    infer_hgvs_c: bool = True
    all: bool = False  # print all possible hgvs_c (for all UTA transcripts)
    show_reverse_strand: bool = (
        False  # show the reverse strand sequence for the chromosome
    )
    alt_aln_method: str = "splign"
    reverse_display: bool = True  # display reverse strand as the main strand (5'->3')



from bioutils.sequences import reverse_complement

from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class ChromReverseSeqRendered(BasicRenderer):
    """
    ChromReverseSeqRendered is a renderer class that extends BasicRenderer to provide
    functionality for rendering reverse complement sequences with optional color coding.
    Methods
    -------
    legend() -> str:
        Returns a string representing the legend for the sequence orientation.
    display(data: VariantData) -> str:
        Colors the reference sequences with adenine (A, green), thymine (T, red), cytosine (C, yellow),
        and guanine (G, blue) based on the provided VariantData. The sequences are displayed in reverse
        complement form.
    """

    def legend(self) -> str:
        if self.orientation < 0 and self.config.reverse_display:
            arrow = "->"
        else:
            arrow = "<-"

        return f"seq    {arrow} : "

    def display(self, data: VariantData) -> str:
        """colors the ref sequences with adenine (A, green), thymine (T, red), cytosine (C, yellow), and guanine (G, blue)"""
        from hgvs.pretty.console.constants import ENDC, COLOR_MAP

        var_seq = ""
        for p in data.position_details:
            c = reverse_complement(p.ref)

            if not c:
                var_seq += "."
                continue

            if self.config.use_color:
                var_seq += COLOR_MAP.get(c, COLOR_MAP["N"])

            var_seq += c

            if self.config.use_color:
                var_seq += ENDC

        return var_seq



# -*- coding: utf-8 -*-
import os
import unittest

import pytest
from support import CACHE

import hgvs
import hgvs.dataproviders.uta
import hgvs.location
import hgvs.parser
from hgvs.alignmentmapper import AlignmentMapper
from hgvs.exceptions import HGVSDataNotAvailableError, HGVSInvalidIntervalError


@pytest.mark.quick
class Test_AlignmentMapper(unittest.TestCase):
    ref = "GRCh37.p10"

    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        cls.parser = hgvs.parser.Parser()

    def test_alignmentmapper_failures(self):
        hgvs.global_config.mapping.strict_bounds = True
        with self.assertRaises(HGVSDataNotAvailableError):
            AlignmentMapper(self.hdp, tx_ac="bogus", alt_ac="NM_033089.6", alt_aln_method="splign")
        with self.assertRaises(HGVSDataNotAvailableError):
            AlignmentMapper(
                self.hdp, tx_ac="bogus", alt_ac="NM_033089.6", alt_aln_method="transcript"
            )
        with self.assertRaises(HGVSDataNotAvailableError):
            AlignmentMapper(self.hdp, tx_ac="NM_033089.6", alt_ac="bogus", alt_aln_method="splign")
        with self.assertRaises(HGVSDataNotAvailableError):
            AlignmentMapper(
                self.hdp, tx_ac="NM_000051.3", alt_ac="NC_000011.9", alt_aln_method="bogus"
            )
        with self.assertRaises(HGVSInvalidIntervalError):
            AlignmentMapper(self.hdp, "NM_000348.3", "NC_000002.11", "splign").n_to_c(
                self.parser.parse_n_interval("-1")
            )
        with self.assertRaises(HGVSInvalidIntervalError):
            AlignmentMapper(self.hdp, "NM_000348.3", "NC_000002.11", "splign").c_to_n(
                self.parser.parse_c_interval("99999")
            )

    def x_test_alignmentmapper_AlignmentMapper_LCE3C_uncertain(self):
        # ? is not yet supported
        """Use NM_178434.2 tests to test mapping with uncertain positions"""
        tx_ac = "NM_178434.2"
        alt_ac = "NC_000001.10"
        test_cases = [
            {
                "g": self.parser.parse_g_interval("(?_152573139)"),
                "n": self.parser.parse_n_interval("(?_2)"),
                "c": self.parser.parse_c_interval("(?_-69)"),
            },
            {
                "g": self.parser.parse_g_interval("(152573138_?)"),
                "n": self.parser.parse_n_interval("(1_?)"),
                "c": self.parser.parse_c_interval("(-70_?)"),
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def test_alignmentmapper_AlignmentMapper_LCE3C(self):
        """NM_178434.2: LCE3C single exon, strand = +1, all coordinate input/output are in HGVS"""
        tx_ac = "NM_178434.2"
        alt_ac = "NC_000001.10"
        test_cases = [
            # 5'
            {
                "g": self.parser.parse_g_interval("152573138"),
                "n": self.parser.parse_n_interval("1"),
                "c": self.parser.parse_c_interval("-70"),
            },
            {
                "g": self.parser.parse_g_interval("152573140"),
                "n": self.parser.parse_n_interval("3"),
                "c": self.parser.parse_c_interval("-68"),
            },
            # cds
            {
                "g": self.parser.parse_g_interval("152573207"),
                "n": self.parser.parse_n_interval("70"),
                "c": self.parser.parse_c_interval("-1"),
            },
            {
                "g": self.parser.parse_g_interval("152573208"),
                "n": self.parser.parse_n_interval("71"),
                "c": self.parser.parse_c_interval("1"),
            },
            # 3'
            {
                "g": self.parser.parse_g_interval("152573492"),
                "n": self.parser.parse_n_interval("355"),
                "c": self.parser.parse_c_interval("285"),
            },
            {
                "g": self.parser.parse_g_interval("152573493"),
                "n": self.parser.parse_n_interval("356"),
                "c": self.parser.parse_c_interval("*1"),
            },
            {
                "g": self.parser.parse_g_interval("152573560"),
                "n": self.parser.parse_n_interval("423"),
                "c": self.parser.parse_c_interval("*68"),
            },
            {
                "g": self.parser.parse_g_interval("152573562"),
                "n": self.parser.parse_n_interval("425"),
                "c": self.parser.parse_c_interval("*70"),
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def test_alignmentmapper_AlignmentMapper_HIST3H2A(self):
        """NM_033445.2: LCE3C single exon, strand = -1, all coordinate input/output are in HGVS"""
        tx_ac = "NM_033445.2"
        alt_ac = "NC_000001.10"
        test_cases = [
            # 3'
            {
                "g": self.parser.parse_g_interval("228645560"),
                "n": self.parser.parse_n_interval("1"),
                "c": self.parser.parse_c_interval("-42"),
            },
            {
                "g": self.parser.parse_g_interval("228645558"),
                "n": self.parser.parse_n_interval("3"),
                "c": self.parser.parse_c_interval("-40"),
            },
            # cds
            {
                "g": self.parser.parse_g_interval("228645519"),
                "n": self.parser.parse_n_interval("42"),
                "c": self.parser.parse_c_interval("-1"),
            },
            {
                "g": self.parser.parse_g_interval("228645518"),
                "n": self.parser.parse_n_interval("43"),
                "c": self.parser.parse_c_interval("1"),
            },
            # 5'
            {
                "g": self.parser.parse_g_interval("228645126"),
                "n": self.parser.parse_n_interval("435"),
                "c": self.parser.parse_c_interval("393"),
            },
            {
                "g": self.parser.parse_g_interval("228645125"),
                "n": self.parser.parse_n_interval("436"),
                "c": self.parser.parse_c_interval("*1"),
            },
            {
                "g": self.parser.parse_g_interval("228645124"),
                "n": self.parser.parse_n_interval("437"),
                "c": self.parser.parse_c_interval("*2"),
            },
            {
                "g": self.parser.parse_g_interval("228645065"),
                "n": self.parser.parse_n_interval("496"),
                "c": self.parser.parse_c_interval("*61"),
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def test_alignmentmapper_AlignmentMapper_LCE2B(self):
        """NM_014357.4: LCE2B, two exons, strand = +1, all coordinate input/output are in HGVS"""
        tx_ac = "NM_014357.4"
        alt_ac = "NC_000001.10"
        test_cases = [
            # 5'
            {
                "g": self.parser.parse_g_interval("152658599"),
                "n": self.parser.parse_n_interval("1"),
                "c": self.parser.parse_c_interval("-54"),
            },
            {
                "g": self.parser.parse_g_interval("152658601"),
                "n": self.parser.parse_n_interval("3"),
                "c": self.parser.parse_c_interval("-52"),
            },
            # cds
            {
                "g": self.parser.parse_g_interval("152659319"),
                "n": self.parser.parse_n_interval("54"),
                "c": self.parser.parse_c_interval("-1"),
            },
            {
                "g": self.parser.parse_g_interval("152659320"),
                "n": self.parser.parse_n_interval("55"),
                "c": self.parser.parse_c_interval("1"),
            },
            # around end of exon 1
            {
                "g": self.parser.parse_g_interval("152658632"),
                "n": self.parser.parse_n_interval("34"),
                "c": self.parser.parse_c_interval("-21"),
            },
            {
                "g": self.parser.parse_g_interval("152658633"),
                "n": self.parser.parse_n_interval("34+1"),
                "c": self.parser.parse_c_interval("-21+1"),
            },
            # span
            {
                "g": self.parser.parse_g_interval("152658633_152659299"),
                "n": self.parser.parse_n_interval("34+1_35-1"),
                "c": self.parser.parse_c_interval("-21+1_-20-1"),
            },
            # around beginning of exon 2
            {
                "g": self.parser.parse_g_interval("152659300"),
                "n": self.parser.parse_n_interval("35"),
                "c": self.parser.parse_c_interval("-20"),
            },
            {
                "g": self.parser.parse_g_interval("152659299"),
                "n": self.parser.parse_n_interval("35-1"),
                "c": self.parser.parse_c_interval("-20-1"),
            },
            # around end of exon 2
            {
                "g": self.parser.parse_g_interval("152659652"),
                "n": self.parser.parse_n_interval("387"),
                "c": self.parser.parse_c_interval("333"),
            },
            {
                "g": self.parser.parse_g_interval("152659653"),
                "n": self.parser.parse_n_interval("388"),
                "c": self.parser.parse_c_interval("*1"),
            },
            # span
            {
                "g": self.parser.parse_g_interval("152659651_152659654"),
                "n": self.parser.parse_n_interval("386_389"),
                "c": self.parser.parse_c_interval("332_*2"),
            },
            # 3'
            {
                "g": self.parser.parse_g_interval("152659877"),
                "n": self.parser.parse_n_interval("612"),
                "c": self.parser.parse_c_interval("*225"),
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def test_alignmentmapper_AlignmentMapper_PTH2(self):
        """NM_178449.3: PTH2, two exons, strand = -1, all coordinate input/output are in HGVS"""
        tx_ac = "NM_178449.3"
        alt_ac = "NC_000019.9"
        test_cases = [
            # 3'
            {
                "g": self.parser.parse_g_interval("49926698"),
                "n": self.parser.parse_n_interval("1"),
                "c": self.parser.parse_c_interval("-102"),
            },
            # cds
            {
                "g": self.parser.parse_g_interval("49926597"),
                "n": self.parser.parse_n_interval("102"),
                "c": self.parser.parse_c_interval("-1"),
            },
            {
                "g": self.parser.parse_g_interval("49926596"),
                "n": self.parser.parse_n_interval("103"),
                "c": self.parser.parse_c_interval("1"),
            },
            # around end of exon 1
            {
                "g": self.parser.parse_g_interval("49926469"),
                "n": self.parser.parse_n_interval("230"),
                "c": self.parser.parse_c_interval("128"),
            },
            {
                "g": self.parser.parse_g_interval("49926468"),
                "n": self.parser.parse_n_interval("230+1"),
                "c": self.parser.parse_c_interval("128+1"),
            },
            # span
            {
                "g": self.parser.parse_g_interval("49925901_49926467"),
                "n": self.parser.parse_n_interval("230+2_231-2"),
                "c": self.parser.parse_c_interval("128+2_129-2"),
            },
            # around beginning of exon 2
            {
                "g": self.parser.parse_g_interval("49925900"),
                "n": self.parser.parse_n_interval("231-1"),
                "c": self.parser.parse_c_interval("129-1"),
            },
            {
                "g": self.parser.parse_g_interval("49925899"),
                "n": self.parser.parse_n_interval("231"),
                "c": self.parser.parse_c_interval("129"),
            },
            # around end of exon 2
            {
                "g": self.parser.parse_g_interval("49925725"),
                "n": self.parser.parse_n_interval("405"),
                "c": self.parser.parse_c_interval("303"),
            },
            {
                "g": self.parser.parse_g_interval("49925724"),
                "n": self.parser.parse_n_interval("406"),
                "c": self.parser.parse_c_interval("*1"),
            },
            {
                "g": self.parser.parse_g_interval("49925671"),
                "n": self.parser.parse_n_interval("459"),
                "c": self.parser.parse_c_interval("*54"),
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def run_cases(self, tx_ac, alt_ac, test_cases):
        am = AlignmentMapper(self.hdp, tx_ac, alt_ac, alt_aln_method="splign")
        for test_case in test_cases:
            assert test_case["c"] == am.g_to_c(
                test_case["g"]
            ), f"{tx_ac}~{alt_ac} {test_case['c']} am.g_to_c failed"
            assert test_case["c"] == am.n_to_c(
                test_case["n"]
            ), f"{tx_ac}~{alt_ac} {test_case['c']} am.n_to_c failed"
            assert test_case["g"] == am.c_to_g(
                test_case["c"]
            ), f"{tx_ac}~{alt_ac} {test_case['g']} am.c_to_g failed"
            assert test_case["g"] == am.n_to_g(
                test_case["n"]
            ), f"{tx_ac}~{alt_ac} {test_case['g']} am.n_to_g failed"
            assert test_case["n"] == am.c_to_n(
                test_case["c"]
            ), f"{tx_ac}~{alt_ac} {test_case['n']} am.c_to_n failed"
            assert test_case["n"] == am.g_to_n(
                test_case["g"]
            ), f"{tx_ac}~{alt_ac} {test_case['n']} am.g_to_n failed"


if __name__ == "__main__":
    unittest.main()


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Projects variants between sequences using AlignmentMapper.

"""

import copy
import logging

from bioutils.sequences import TranslationTable, reverse_complement

import hgvs
import hgvs.alignmentmapper
import hgvs.edit
import hgvs.location
import hgvs.normalizer
import hgvs.posedit
import hgvs.sequencevariant
import hgvs.validator
from hgvs.decorators.lru_cache import lru_cache
from hgvs.enums import PrevalidationLevel
from hgvs.exceptions import HGVSInvalidVariantError, HGVSUnsupportedOperationError
from hgvs.utils import altseq_to_hgvsp, altseqbuilder
from hgvs.utils.position import get_start_end, get_start_end_interbase
from hgvs.utils.reftranscriptdata import RefTranscriptData

_logger = logging.getLogger(__name__)


class VariantMapper:
    r"""Maps SequenceVariant objects between g., n., r., c., and p. representations.

    g⟷{c,n,r} projections are similar in that c, n, and r variants
    may use intronic coordinates. There are two essential differences
    that distinguish the three types:

    * Sequence start: In n and r variants, position 1 is the sequence
      start; in c variants, 1 is the transcription start site.
    * Alphabet: In n and c variants, sequences are DNA; in
      r. variants, sequences are RNA.

    This differences are summarized in this diagram::

      g ----acgtatgcac--gtctagacgt----      ----acgtatgcac--gtctagacgt----      ----acgtatgcac--gtctagacgt----
            \         \/         /              \         \/         /              \         \/         /
      c      acgtATGCACGTCTAGacgt         n      acgtatgcacgtctagacgt         r      acguaugcacgucuagacgu
                 1                               1                                   1
      p          MetHisValTer

    The g excerpt and exon structures are identical. The g⟷n
    transformation, which is the most basic, accounts for the offset
    of the aligned sequences (shown with "1") and the exon structure.
    The g⟷c transformation is akin to g⟷n transformation, but
    requires an addition offset to account for the translation start
    site (c.1).  The CDS in uppercase. The g⟷r transformation is
    akin to g⟷n transformation with a change of alphabet.

    Therefore, this this code uses g⟷n as the core transformation
    between genomic and c, n, and r variants: All c⟷g and r⟷g
    transformations use n⟷g after accounting for the above
    differences. For example, c_to_g accounts for the transcription
    start site offset, then calls n_to_g.

    All methods require and return objects of type
    :class:`hgvs.sequencevariant.SequenceVariant`.

    """

    def __init__(
        self,
        hdp,
        replace_reference=hgvs.global_config.mapping.replace_reference,
        prevalidation_level=hgvs.global_config.mapping.prevalidation_level,
        add_gene_symbol=hgvs.global_config.mapping.add_gene_symbol,
    ):
        """
        :param bool replace_reference: replace reference (entails additional network access)
        :param str prevalidation_level: None or Intrinsic or Extrinsic validation before mapping

        """
        self.hdp = hdp
        self.replace_reference = replace_reference
        self.add_gene_symbol = add_gene_symbol
        if prevalidation_level is None:
            self.prevalidation_level = PrevalidationLevel.NONE
        else:
            self.prevalidation_level = PrevalidationLevel[prevalidation_level.upper()]
        if self.prevalidation_level == PrevalidationLevel.NONE:
            self._validator = None
        elif self.prevalidation_level == PrevalidationLevel.INTRINSIC:
            self._validator = hgvs.validator.IntrinsicValidator(strict=False)
        else:
            self._validator = hgvs.validator.Validator(self.hdp, strict=False)
        self.left_normalizer = hgvs.normalizer.Normalizer(
            hdp, shuffle_direction=5, variantmapper=self
        )

    # ############################################################################
    # g⟷t
    def g_to_t(self, var_g, tx_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        if var_g.type not in "gm":
            raise HGVSInvalidVariantError("Expected a g. or m. variant; got " + str(var_g))
        if self._validator:
            self._validator.validate(var_g)
        var_g.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method
        )
        if mapper.is_coding_transcript:
            var_out = VariantMapper.g_to_c(
                self, var_g=var_g, tx_ac=tx_ac, alt_aln_method=alt_aln_method
            )
        else:
            var_out = VariantMapper.g_to_n(
                self, var_g=var_g, tx_ac=tx_ac, alt_aln_method=alt_aln_method
            )
        return var_out

    def t_to_g(self, var_t, alt_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        if var_t.type not in "cn":
            raise HGVSInvalidVariantError("Expected a c. or n. variant; got " + str(var_t))
        if self._validator:
            self._validator.validate(var_t)
        var_t.fill_ref(self.hdp, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        if var_t.type == "c":
            var_out = VariantMapper.c_to_g(
                self, var_c=var_t, alt_ac=alt_ac, alt_aln_method=alt_aln_method
            )
        else:
            var_out = VariantMapper.n_to_g(
                self, var_n=var_t, alt_ac=alt_ac, alt_aln_method=alt_aln_method
            )
        return var_out

    # ############################################################################
    # g⟷n
    def g_to_n(self, var_g, tx_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed g. variant, return a n. variant on the specified
        transcript using the specified alignment method (default is
        "splign" from NCBI).

        :param hgvs.sequencevariant.SequenceVariant var_g: a variant object
        :param str tx_ac: a transcript accession (e.g., NM_012345.6 or ENST012345678)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`) using transcript (n.) coordinates
        :raises HGVSInvalidVariantError: if var_g is not of type "g"

        """

        if var_g.type not in "gm":
            raise HGVSInvalidVariantError("Expected a g. or m. variant; got " + str(var_g))
        if self._validator:
            self._validator.validate(var_g)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method
        )

        if (
            mapper.strand == -1
            and not hgvs.global_config.mapping.strict_bounds
            and not mapper.g_interval_is_inbounds(var_g.posedit.pos)
        ):
            _logger.info("Renormalizing out-of-bounds minus strand variant on genomic sequence")
            var_g = self.left_normalizer.normalize(var_g)

        var_g.fill_ref(self.hdp)
        pos_n = mapper.g_to_n(var_g.posedit.pos)
        if not pos_n.uncertain:
            edit_n = self._convert_edit_check_strand(mapper.strand, var_g.posedit.edit)
            if (
                edit_n.type == "ins"
                and pos_n.start.offset == 0
                and pos_n.end.offset == 0
                and pos_n.end - pos_n.start > 1
            ):
                pos_n.start.base += 1
                pos_n.end.base -= 1
                edit_n.ref = ""
        else:
            # variant at alignment gap
            pos_g = mapper.n_to_g(pos_n)
            edit_n = hgvs.edit.NARefAlt(
                ref="", alt=self._get_altered_sequence(mapper.strand, pos_g, var_g)
            )
        pos_n.uncertain = var_g.posedit.pos.uncertain
        var_n = hgvs.sequencevariant.SequenceVariant(
            ac=tx_ac, type="n", posedit=hgvs.posedit.PosEdit(pos_n, edit_n)
        )
        if (
            self.replace_reference
            and var_n.posedit.pos.start.base >= 0
            and var_n.posedit.pos.end.base < mapper.tgt_len
        ):
            self._replace_reference(var_n)
        if self.add_gene_symbol:
            self._update_gene_symbol(var_n, var_g.gene)
        return var_n

    def n_to_g(self, var_n, alt_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed n. variant, return a g. variant on the specified
        transcript using the specified alignment method (default is
        "splign" from NCBI).

        :param hgvs.sequencevariant.SequenceVariant var_n: a variant object
        :param str alt_ac: a reference sequence accession (e.g., NC_000001.11)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_n is not of type "n"

        """

        if not (var_n.type == "n"):
            raise HGVSInvalidVariantError("Expected a n. variant; got " + str(var_n))
        if self._validator:
            self._validator.validate(var_n)
        var_n.fill_ref(self.hdp, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_n.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method
        )
        pos_g = mapper.n_to_g(var_n.posedit.pos)
        if not pos_g.uncertain:
            edit_g = self._convert_edit_check_strand(mapper.strand, var_n.posedit.edit)
            if edit_g.type == "ins" and pos_g.end - pos_g.start > 1:
                pos_g.start.base += 1
                pos_g.end.base -= 1
                edit_g.ref = ""
        else:
            # variant at alignment gap
            pos_n = mapper.g_to_n(pos_g)
            edit_g = hgvs.edit.NARefAlt(
                ref="", alt=self._get_altered_sequence(mapper.strand, pos_n, var_n)
            )
        pos_g.uncertain = var_n.posedit.pos.uncertain
        var_g = hgvs.sequencevariant.SequenceVariant(
            ac=alt_ac, type="g", posedit=hgvs.posedit.PosEdit(pos_g, edit_g)
        )
        if self.replace_reference:
            self._replace_reference(var_g, alt_ac, alt_aln_method)
        # No gene symbol for g. variants (actually, *should* for NG, but no way to distinguish)
        return var_g

    # ############################################################################
    # g⟷c
    def g_to_c(self, var_g, tx_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed g. variant, return a c. variant on the specified
        transcript using the specified alignment method (default is
        "splign" from NCBI).

        :param hgvs.sequencevariant.SequenceVariant var_g: a variant object
        :param str tx_ac: a transcript accession (e.g., NM_012345.6 or ENST012345678)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`) using CDS coordinates
        :raises HGVSInvalidVariantError: if var_g is not of type "g"

        """

        if var_g.type not in "gm":
            raise HGVSInvalidVariantError("Expected a g. or m. variant; got " + str(var_g))
        if self._validator:
            self._validator.validate(var_g)
        var_g.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method
        )
        pos_c = mapper.g_to_c(var_g.posedit.pos)
        if not pos_c.uncertain:
            edit_c = self._convert_edit_check_strand(mapper.strand, var_g.posedit.edit)
            if (
                edit_c.type == "ins"
                and pos_c.start.offset == 0
                and pos_c.end.offset == 0
                and pos_c.start.datum == pos_c.end.datum
                and pos_c.end - pos_c.start > 1
            ):
                pos_c.start.base += 1
                pos_c.end.base -= 1
                edit_c.ref = ""
        else:
            # variant at alignment gap
            pos_g = mapper.c_to_g(pos_c)
            edit_c = hgvs.edit.NARefAlt(
                ref="", alt=self._get_altered_sequence(mapper.strand, pos_g, var_g)
            )
        pos_c.uncertain = var_g.posedit.pos.uncertain
        var_c = hgvs.sequencevariant.SequenceVariant(
            ac=tx_ac, type="c", posedit=hgvs.posedit.PosEdit(pos_c, edit_c)
        )
        if self.replace_reference:
            self._replace_reference(var_c)
        if self.add_gene_symbol:
            self._update_gene_symbol(var_c, var_g.gene)
        return var_c

    def c_to_g(self, var_c, alt_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed c. variant, return a g. variant on the specified
        transcript using the specified alignment method (default is
        "splign" from NCBI).

        :param hgvs.sequencevariant.SequenceVariant var_c: a variant object
        :param str alt_ac: a reference sequence accession (e.g., NC_000001.11)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_c is not of type "c"

        """

        if not (var_c.type == "c"):
            raise HGVSInvalidVariantError("Expected a cDNA (c.); got " + str(var_c))
        if self._validator:
            self._validator.validate(var_c)
        var_c.fill_ref(self.hdp, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_c.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method
        )
        pos_g = mapper.c_to_g(var_c.posedit.pos)
        if not pos_g.uncertain:
            edit_g = self._convert_edit_check_strand(mapper.strand, var_c.posedit.edit)
            if edit_g.type == "ins" and pos_g.end - pos_g.start > 1:
                pos_g.start.base += 1
                pos_g.end.base -= 1
                edit_g.ref = ""
        else:
            # variant at alignment gap
            var_n = copy.deepcopy(var_c)
            var_n.posedit.pos = mapper.c_to_n(var_c.posedit.pos)
            var_n.type = "n"
            pos_n = mapper.g_to_n(pos_g)
            edit_g = hgvs.edit.NARefAlt(
                ref="", alt=self._get_altered_sequence(mapper.strand, pos_n, var_n)
            )
        pos_g.uncertain = var_c.posedit.pos.uncertain
        var_g = hgvs.sequencevariant.SequenceVariant(
            ac=alt_ac, type="g", posedit=hgvs.posedit.PosEdit(pos_g, edit_g)
        )
        if self.replace_reference:
            self._replace_reference(var_g, alt_ac, alt_aln_method)
        return var_g

    # ############################################################################
    # c⟷n
    def c_to_n(self, var_c, alt_ac=None, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed c. variant, return a n. variant on the specified
        transcript using the specified alignment method (default is
        "transcript" indicating a self alignment).

        :param hgvs.sequencevariant.SequenceVariant var_c: a variant object
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_c is not of type "c"

        """

        if not (var_c.type == "c"):
            raise HGVSInvalidVariantError("Expected a cDNA (c.); got " + str(var_c))
        if self._validator:
            self._validator.validate(var_c)
        var_c.fill_ref(self.hdp, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_c.ac, alt_ac=var_c.ac, alt_aln_method="transcript"
        )
        pos_n = mapper.c_to_n(var_c.posedit.pos)
        if isinstance(var_c.posedit.edit, hgvs.edit.NARefAlt | hgvs.edit.Dup | hgvs.edit.Inv):
            edit_n = copy.deepcopy(var_c.posedit.edit)
        else:
            msg = "Only NARefAlt/Dup/Inv types are currently implemented"
            raise HGVSUnsupportedOperationError(msg)
        var_n = hgvs.sequencevariant.SequenceVariant(
            ac=var_c.ac, type="n", posedit=hgvs.posedit.PosEdit(pos_n, edit_n)
        )
        if self.replace_reference:
            self._replace_reference(var_n, alt_ac, alt_aln_method)
        if self.add_gene_symbol:
            self._update_gene_symbol(var_n, var_c.gene)
        return var_n

    def n_to_c(self, var_n, alt_ac=None, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed n. variant, return a c. variant on the specified
        transcript using the specified alignment method (default is
        "transcript" indicating a self alignment).

        :param hgvs.sequencevariant.SequenceVariant var_n: a variant object
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_n is not of type "n"

        """

        if not (var_n.type == "n"):
            raise HGVSInvalidVariantError("Expected n. variant; got " + str(var_n))
        if self._validator:
            self._validator.validate(var_n)
        var_n.fill_ref(self.hdp, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_n.ac, alt_ac=var_n.ac, alt_aln_method="transcript"
        )
        pos_c = mapper.n_to_c(var_n.posedit.pos)
        if isinstance(var_n.posedit.edit, hgvs.edit.NARefAlt | hgvs.edit.Dup | hgvs.edit.Inv):
            edit_c = copy.deepcopy(var_n.posedit.edit)
        else:
            msg = "Only NARefAlt/Dup/Inv types are currently implemented"
            raise HGVSUnsupportedOperationError(msg)
        var_c = hgvs.sequencevariant.SequenceVariant(
            ac=var_n.ac, type="c", posedit=hgvs.posedit.PosEdit(pos_c, edit_c)
        )
        if self.replace_reference:
            self._replace_reference(var_c, alt_ac, alt_aln_method)
        if self.add_gene_symbol:
            self._update_gene_symbol(var_c, var_n.gene)
        return var_c

    # ############################################################################
    # c ⟶ p
    def c_to_p(self, var_c, pro_ac=None, alt_ac=None, alt_aln_method=hgvs.global_config.mapping.alt_aln_method, translation_table=TranslationTable.standard):
        """
        Converts a c. SequenceVariant to a p. SequenceVariant on the specified protein accession
        Author: Rudy Rico

        :param SequenceVariant var_c: hgvsc tag
        :param str pro_ac: protein accession
        :rtype: hgvs.sequencevariant.SequenceVariant

        """

        if not (var_c.type == "c"):
            raise HGVSInvalidVariantError("Expected a cDNA (c.) variant; got " + str(var_c))
        if self._validator:
            self._validator.validate(var_c)
        var_c.fill_ref(self.hdp, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        reference_data = RefTranscriptData(
            self.hdp, var_c.ac, pro_ac, translation_table=translation_table
        )
        # Use the actual translation table from reference_data (may have been auto-detected)
        builder = altseqbuilder.AltSeqBuilder(
            var_c, reference_data, translation_table=reference_data.translation_table
        )

        # TODO: handle case where you get 2+ alt sequences back;
        # currently get list of 1 element loop structure implemented
        # to handle this, but doesn't really do anything currently.
        all_alt_data = builder.build_altseq()

        var_ps = []
        for alt_data in all_alt_data:
            builder = altseq_to_hgvsp.AltSeqToHgvsp(reference_data, alt_data)
            var_p = builder.build_hgvsp()
            var_ps.append(var_p)

        var_p = var_ps[0]

        if self.add_gene_symbol:
            self._update_gene_symbol(var_p, var_c.gene)

        return var_p

    ############################################################################
    # Internal methods

    def _replace_reference(  # noqa: PLR0912
        self, var, alt_ac=None, alt_aln_method=hgvs.global_config.mapping.alt_aln_method
    ):
        """fetch reference sequence for variant and update (in-place) if necessary"""

        if var.type not in "cgmnr":
            msg = "Can only update references for type c, g, m, n, r"
            raise HGVSUnsupportedOperationError(msg)

        if var.posedit.edit.type in ("ins", "con"):
            # these types have no reference sequence (zero-width), so return as-is
            return var

        pos = var.posedit.pos
        mapper = None
        if (isinstance(pos.start, hgvs.location.BaseOffsetPosition) and pos.start.offset != 0) or (
            isinstance(pos.end, hgvs.location.BaseOffsetPosition) and pos.end.offset != 0
        ):
            if var.type == "r":
                _logger.info("Can't update reference sequence for intronic variant %s", var)
                return var
            if alt_ac is None:
                _logger.info("Can't update reference sequence for intronic variant %s without alt_ac", var)
                return var
            if var.type == "c":
                mapper = self._fetch_AlignmentMapper(
                    tx_ac=var.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method
                )
                pos = mapper.c_to_g(var.posedit.pos)
                ac = alt_ac
                _type = "g"
            if var.type == "n":
                mapper = self._fetch_AlignmentMapper(
                    tx_ac=var.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method
                )
                pos = mapper.n_to_g(var.posedit.pos)
                ac = alt_ac
                _type = "g"
        elif var.type == "c":
            # For c. variants, we need coords on underlying sequences
            mapper = self._fetch_AlignmentMapper(
                tx_ac=var.ac, alt_ac=var.ac, alt_aln_method="transcript"
            )
            pos = mapper.c_to_n(var.posedit.pos)
            ac = var.ac
            _type = var.type
        else:
            pos = var.posedit.pos
            ac = var.ac
            _type = var.type

        seq_start, seq_end = get_start_end_interbase(pos, outer_confidence=True)
        # When strict_bounds is False and an error occurs, return
        # variant as-is

        if seq_start is None or seq_end is None or seq_start < 0:
            # this is an out-of-bounds variant
            return var

        seq = self.hdp.get_seq(ac, seq_start, seq_end)

        if len(seq) != seq_end - seq_start:
            # tried to read beyond seq end; this is an out-of-bounds variant
            return var

        if _type == "g" and mapper and mapper.strand == -1:
            seq = reverse_complement(seq)

        edit = var.posedit.edit
        if edit.ref != seq:
            _logger.debug("Replaced reference sequence in %s with %s", var, seq)
            edit.ref = seq

        return var

    @lru_cache(maxsize=hgvs.global_config.lru_cache.maxsize)
    def _fetch_AlignmentMapper(self, tx_ac, alt_ac, alt_aln_method):
        """
        Get a new AlignmentMapper for the given transcript accession (ac),
        possibly caching the result.
        """
        return hgvs.alignmentmapper.AlignmentMapper(
            self.hdp, tx_ac=tx_ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method
        )

    @staticmethod
    def _convert_edit_check_strand(strand, edit_in):
        """
        Convert an edit from one type to another, based on the stand and type
        """
        if isinstance(edit_in, hgvs.edit.NARefAlt):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                try:
                    # if smells like an int, do nothing
                    # TODO: should use ref_n, right?
                    int(edit_in.ref)
                    ref = edit_in.ref
                except (ValueError, TypeError):
                    ref = reverse_complement(edit_in.ref)
                edit_out = hgvs.edit.NARefAlt(
                    ref=ref,
                    alt=reverse_complement(edit_in.alt),
                )
        elif isinstance(edit_in, hgvs.edit.Dup):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                edit_out = hgvs.edit.Dup(ref=reverse_complement(edit_in.ref))
        elif isinstance(edit_in, hgvs.edit.Inv):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                try:
                    int(edit_in.ref)
                    ref = edit_in.ref
                except (ValueError, TypeError):
                    ref = reverse_complement(edit_in.ref)
                edit_out = hgvs.edit.Inv(ref=ref)
        else:
            msg = "Only NARefAlt/Dup/Inv types are currently implemented"
            raise NotImplementedError(msg)
        return edit_out

    def _get_altered_sequence(self, strand, interval, var):
        seq = list(self.hdp.get_seq(var.ac, interval.start.base - 1, interval.end.base))
        # positions are 0-based and half-open
        pos_start = var.posedit.pos.start.base - interval.start.base
        pos_end = var.posedit.pos.end.base - interval.start.base + 1
        edit = var.posedit.edit

        if edit.type == "sub":
            seq[pos_start] = edit.alt
        elif edit.type == "del":
            del seq[pos_start:pos_end]
        elif edit.type == "ins":
            seq.insert(pos_start + 1, edit.alt)
        elif edit.type == "delins":
            del seq[pos_start:pos_end]
            seq.insert(pos_start, edit.alt)
        elif edit.type == "dup":
            seq.insert(pos_end, "".join(seq[pos_start:pos_end]))
        elif edit.type == "inv":
            seq[pos_start:pos_end] = list(reverse_complement("".join(seq[pos_start:pos_end])))
        elif edit.type == "identity":
            pass
        else:
            msg = f"Getting altered sequence for {edit.type} is unsupported"
            raise HGVSUnsupportedOperationError(msg)

        seq = "".join(seq)
        if strand == -1:
            seq = reverse_complement(seq)
        return seq

    def _update_gene_symbol(self, var, symbol):
        if not symbol:
            symbol = self.hdp.get_tx_identity_info(var.ac).get("hgnc", None)
        var.gene = symbol
        return var


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



#!/usr/bin/env python
# This generates the hgvs.parser.Parser Python code from the OMeta grammar file
# We generate it offline then keep the statically generated file to reduce startup time
# @see https://github.com/biocommons/hgvs/issues/661

import hashlib
import os
import sys

import parsley
from ometa.grammar import OMeta
from ometa.builder import writePython

if __name__ == "__main__":
    script_path = os.path.realpath(__file__)
    script_dir = os.path.dirname(script_path)
    hgvs_base_dir = os.path.dirname(script_dir)
    grammar_file = os.path.join(hgvs_base_dir, "src/hgvs/_data/hgvs.pymeta")
    generated_code_dir = os.path.join(hgvs_base_dir, "src/hgvs/generated")

    grammar_hash = hashlib.md5(open(grammar_file, 'rb').read()).hexdigest()
    prefix_length = len(hgvs_base_dir) + 1  # extra to also remove slash

    header_template = """# --------------------------------------------------
# THIS IS A GENERATED FILE. DO NOT MODIFY.
# Changes will be overwritten by the generation script.
#    Generated by: {generate_script}
#    Grammar file: {grammar_file}
#    Grammar hash: {grammar_hash}
# Parsley version: {parsley_version}
#  Python version: {python_version}
# --------------------------------------------------
"""
    header = header_template.format(generate_script=script_path[prefix_length:],
                                    grammar_file=grammar_file[prefix_length:],
                                    grammar_hash=grammar_hash,
                                    parsley_version=parsley.__version__,
                                    python_version=sys.version)

    g = OMeta(open(grammar_file).read(), name="Grammar")
    tree = g.parseGrammar("Grammar")
    source = writePython(tree, "Grammar")

    # Create a package for the new file for Python2 compatability
    if not os.path.exists(generated_code_dir):
        os.mkdir(generated_code_dir)
        package_filename = os.path.join(generated_code_dir, "__init__.py")
        with open(package_filename, "w") as _:
            pass

    source_filename = os.path.join(generated_code_dir, "hgvs_grammar.py")
    with open(source_filename, "w") as source_file:
        source_file.write(header + "\n")
        source_file.write(source)



# -*- coding: utf-8 -*-
"""start IPython shell with hgvs initialized. Intended to be used for
experimenting, debugging, and generating bug reports."""

import logging
import os

import IPython

header_string = """############################################################################
hgvs-shell -- interactive hgvs
hgvs version: {v}
data provider url: {hdp.url}
schema_version: {sv}
data_version: {dv}
sequences source: {hdp.seqfetcher.source}

The following variables are defined:
* global_config
* hp, parser, hgvs_parser -- Parser instance
* hdp, hgvs_data_provider -- UTA data provider instance
* vm, variant_mapper, hgvs_variant_mapper -- VariantMapper instance
* am37, hgvs_assembly_mapper_37 -- GRCh37 Assembly Mapper instance
* am38, projector, hgvs_assembly_mapper_38 -- GRCh38 Assembly Mapper instances
* hn, normalizer, hgvs_normalizer -- Normalizer instance
* hv, validator, hgvs_validator) -- Validator instance
* pretty -- PrettyPrint instance

The following functions are available:
  * parse, normalize, validate
  * g_to_c, g_to_n, g_to_t,
  * c_to_g, c_to_n, c_to_p,
  * n_to_c, n_to_g,
  * t_to_g,
  * get_relevant_transcripts

When submitting bug reports, include the version header shown above
and use these variables/variable names whenever possible.

"""


def shell():
    logging.basicConfig(level=os.environ.get("HGVS_LOGGING_LEVEL", logging.WARNING))

    from hgvs.easy import (  # noqa: F401; instances; functionalized methods
        __version__,
        am37,
        am38,
        c_to_g,
        c_to_n,
        c_to_p,
        g_to_c,
        g_to_n,
        g_to_t,
        get_relevant_transcripts,
        global_config,
        hdp,
        hgvs_assembly_mapper_37,
        hgvs_assembly_mapper_38,
        hgvs_data_provider,
        hgvs_normalizer,
        hgvs_parser,
        hgvs_validator,
        hgvs_variant_mapper,
        hn,
        hp,
        hv,
        n_to_c,
        n_to_g,
        normalize,
        normalizer,
        parse,
        parser,
        pretty37,
        pretty38,
        projector,
        t_to_g,
        t_to_p,
        validate,
        validator,
        variant_mapper,
        vm,
    )
    from hgvs.utils.context import variant_context_w_alignment  # noqa

    IPython.embed(
        header=header_string.format(
            v=__version__,
            hdp=hdp,
            sv=hdp.schema_version(),
            dv=hdp.data_version(),
        )
    )


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
import os
import unittest

import pytest
from support import CACHE

import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.variantmapper
from hgvs.enums import PrevalidationLevel
from hgvs.exceptions import HGVSInvalidVariantError


@pytest.mark.quick
class Test_VariantMapper(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        cls.am = hgvs.assemblymapper.AssemblyMapper(cls.hdp)
        cls.am37 = hgvs.assemblymapper.AssemblyMapper(cls.hdp, assembly_name="GRCh37")
        cls.hp = hgvs.parser.Parser()

    def test_VariantMapper_quick(self):
        # From garcia.tsv:
        hgvs_g = "NC_000007.13:g.36561662C>T"
        hgvs_c = "NM_001637.3:c.1582G>A"
        hgvs_p = "NP_001628.1:p.(Gly528Arg)"

        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        var_c = self.am.g_to_c(var_g, "NM_001637.3")
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_c), hgvs_c)
        self.assertEqual(str(var_p), hgvs_p)

        with self.assertRaises(HGVSInvalidVariantError):
            self.am.c_to_p(self.hp.parse_hgvs_variant("NM_000059.3:c.7790delAAG"))

        hgvs_c = "NM_000059.3:c.7791A>G"
        hgvs_p = "NP_000050.2:p.(Lys2597=)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

        hgvs_c = "NM_000302.3:c.1594_1596del"
        hgvs_p = "NP_000293.2:p.(Glu532del)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

        hgvs_c = "NM_000090.3:c.2490_2516del"
        hgvs_p = "NP_000081.1:p.(Glu832_Gly840del)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

        hgvs_c = "NM_001292004.1:c.376="
        hgvs_g = "NC_000005.10:g.74715659="

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_g = self.am.c_to_g(var_c)

        self.assertEqual(str(var_g), hgvs_g)

        hgvs_c = "NM_000116.4:c.-8_-3inv6"
        hgvs_g = "NC_000023.11:g.154411836_154411841inv"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_g = self.am.c_to_g(var_c)

        self.assertEqual(str(var_g), hgvs_g)

        hgvs_c = "NM_000348.3:c.89_91inv3"
        hgvs_g = "NC_000002.12:g.31580810_31580812inv"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_g = self.am.c_to_g(var_c)

        self.assertEqual(str(var_g), hgvs_g)

        hgvs_c = "NM_001637.3:c.1582_1583inv"
        hgvs_p = "NP_001628.1:p.(Gly528Pro)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

        hgvs_c = "NM_025137.3:c.-20_*20inv"
        hgvs_p = "NP_079413.3:p.?"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

    def test_projection_at_alignment_discrepancy(self):
        hgvs_g = "NC_000019.10:g.50378563_50378564insTG"
        hgvs_n = "NM_007121.5:n.796_798delinsTG"

        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        var_n = self.am.g_to_n(var_g, "NM_007121.5")

        self.assertEqual(str(var_n), hgvs_n)

        hgvs_g = "NC_000007.14:g.149779575delC"
        hgvs_n = "NM_198455.2:n.1115_1116insAG"

        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        var_n = self.am.g_to_n(var_g, "NM_198455.2")

        self.assertEqual(str(var_n), hgvs_n)

        # issue-353
        hgvs_g = "NC_000012.11:g.122064775C>T"
        hgvs_c = "NM_032790.3:c.127_128insTGCCAC"

        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        var_c = self.am.g_to_c(var_g, "NM_032790.3")

        self.assertEqual(str(var_c), hgvs_c)

        # issue-461
        hgvs_g = "NC_000002.11:g.73675227_73675228insCTC"
        hgvs_c = "NM_015120.4:c.1574_1576="

        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        var_c = self.am37.g_to_c(var_g, "NM_015120.4")

        self.assertEqual(str(var_c), hgvs_c)

        # issue-259
        hgvs_c = "NM_000116.4:c.-120_-119insT"
        hgvs_g = "NC_000023.10:g.153640061="

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_g = self.am37.c_to_g(var_c)

        self.assertEqual(str(var_g), hgvs_g)

        hgvs_c = "NM_000348.3:c.88del"
        hgvs_g = "NC_000002.11:g.31805882_31805883="

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_g = self.am37.c_to_g(var_c)

        self.assertEqual(str(var_g), hgvs_g)

    def test_c_to_p_with_stop_gain(self):
        # issue-474
        hgvs_c = "NM_080877.2:c.1733_1735delinsTTT"
        hgvs_p = "NP_543153.1:p.(Pro578_Lys579delinsLeuTer)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

        # issue-492
        hgvs_c = "NM_001034853.1:c.2847_2848delAGinsCT"
        hgvs_p = "NP_001030025.1:p.(Glu949_Glu950delinsAspTer)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

        hgvs_c = "NM_001034853.1:c.2847_2848inv"
        hgvs_p = "NP_001030025.1:p.(Glu949_Glu950delinsAspTer)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

        hgvs_c = "NM_080877.2:c.1735A>T"
        hgvs_p = "NP_543153.1:p.(Lys579Ter)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)

        hgvs_c = "NM_080877.2:c.1795_*3delinsTAG"
        hgvs_p = "NP_543153.1:p.(Leu599Ter)"

        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.am.c_to_p(var_c)

        self.assertEqual(str(var_p), hgvs_p)


class Test_RefReplacement(unittest.TestCase):
    test_cases = [
        # These casese attempt to test reference update in four dimensions:
        # - variant type: n, c, g
        # - major mapping paths: c<->n, c<->g, n<->g
        # - variant class: sub, del, ins, delins, dup
        # - strand: +/-
        # ADRB2    │ NM_000024.5 │  239 │ 1481 │ NC_000005.9  │  1 │ 148206155,148208197 | 284=1X32=1X1724=
        # cseq = hdp.fetch_seq("NM_000024.5")
        # gseq = hdp.fetch_seq("NC_000005.9",148206155,148208197)
        # cseq[280:290] = "CAATAGAAGC"
        # gseq[280:290] = "CAATGGAAGC"
        #                      ^ @ n.285
        # These variants are in and around the first sub:
        {
            "c": "NM_000024.5:c.42C>N",
            "g": "NC_000005.9:g.148206436C>N",
            "n": "NM_000024.5:n.281C>N",
        },
        {
            "c": "NM_000024.5:c.43A>N",
            "g": "NC_000005.9:g.148206437A>N",
            "n": "NM_000024.5:n.282A>N",
        },
        {
            "c": "NM_000024.5:c.44A>N",
            "g": "NC_000005.9:g.148206438A>N",
            "n": "NM_000024.5:n.283A>N",
        },
        {
            "c": "NM_000024.5:c.45T>N",
            "g": "NC_000005.9:g.148206439T>N",
            "n": "NM_000024.5:n.284T>N",
        },
        {
            "c": "NM_000024.5:c.46A>N",
            "g": "NC_000005.9:g.148206440G>N",
            "n": "NM_000024.5:n.285A>N",
        },  # ref repl
        {
            "c": "NM_000024.5:c.47G>N",
            "g": "NC_000005.9:g.148206441G>N",
            "n": "NM_000024.5:n.286G>N",
        },
        {
            "c": "NM_000024.5:c.48A>N",
            "g": "NC_000005.9:g.148206442A>N",
            "n": "NM_000024.5:n.287A>N",
        },
        {
            "c": "NM_000024.5:c.49A>N",
            "g": "NC_000005.9:g.148206443A>N",
            "n": "NM_000024.5:n.288A>N",
        },
        {
            "c": "NM_000024.5:c.50G>N",
            "g": "NC_000005.9:g.148206444G>N",
            "n": "NM_000024.5:n.289G>N",
        },
        {
            "c": "NM_000024.5:c.51C>N",
            "g": "NC_000005.9:g.148206445C>N",
            "n": "NM_000024.5:n.290C>N",
        },
        # ins, del, delins, dup:
        {
            "c": "NM_000024.5:c.46_47insNN",
            "g": "NC_000005.9:g.148206440_148206441insNN",
            "n": "NM_000024.5:n.285_286insNN",
        },
        {
            "c": "NM_000024.5:c.45_47delTAG",
            "g": "NC_000005.9:g.148206439_148206441delTGG",
            "n": "NM_000024.5:n.284_286delTAG",
        },
        {
            "c": "NM_000024.5:c.45_47delTAGinsNNNN",
            "g": "NC_000005.9:g.148206439_148206441delTGGinsNNNN",
            "n": "NM_000024.5:n.284_286delTAGinsNNNN",
        },
        {
            "c": "NM_000024.5:c.45_47delTAGinsNNNN",
            "g": "NC_000005.9:g.148206439_148206441delTGGinsNNNN",
            "n": "NM_000024.5:n.284_286delTAGinsNNNN",
        },
        {
            "c": "NM_000024.5:c.46dupA",
            "g": "NC_000005.9:g.148206440dupG",
            "n": "NM_000024.5:n.285dupA",
        },
        # IFNA16   │ NM_002173.2 │    6 │  576 │ NC_000009.11 │ -1 │  21216371, 21217310 | 691=2X246=
        # cseq = hdp.fetch_seq("NM_002173.2")
        # gseq = reverse_complement(hdp.fetch_seq("NC_000009.11",21216371,21217310))
        # cseq[685:695] = "AAATTTCAAA"
        # gseq[685:695] = "AAATTTTCAA"
        #                        ^^ @ n.692_693
        # These variants are in and around the 2X substitution
        {
            "c": "NM_002173.2:c.*110A>N",
            "g": "NC_000009.11:g.21216625T>N",
            "n": "NM_002173.2:n.686A>N",
        },
        {
            "c": "NM_002173.2:c.*111A>N",
            "g": "NC_000009.11:g.21216624T>N",
            "n": "NM_002173.2:n.687A>N",
        },
        {
            "c": "NM_002173.2:c.*112A>N",
            "g": "NC_000009.11:g.21216623T>N",
            "n": "NM_002173.2:n.688A>N",
        },
        {
            "c": "NM_002173.2:c.*113T>N",
            "g": "NC_000009.11:g.21216622A>N",
            "n": "NM_002173.2:n.689T>N",
        },
        {
            "c": "NM_002173.2:c.*114T>N",
            "g": "NC_000009.11:g.21216621A>N",
            "n": "NM_002173.2:n.690T>N",
        },
        {
            "c": "NM_002173.2:c.*115T>N",
            "g": "NC_000009.11:g.21216620A>N",
            "n": "NM_002173.2:n.691T>N",
        },
        {
            "c": "NM_002173.2:c.*116C>N",
            "g": "NC_000009.11:g.21216619A>N",
            "n": "NM_002173.2:n.692C>N",
        },  # ref repl
        {
            "c": "NM_002173.2:c.*117A>N",
            "g": "NC_000009.11:g.21216618G>N",
            "n": "NM_002173.2:n.693A>N",
        },  # ref repl
        {
            "c": "NM_002173.2:c.*118A>N",
            "g": "NC_000009.11:g.21216617T>N",
            "n": "NM_002173.2:n.694A>N",
        },
        {
            "c": "NM_002173.2:c.*119A>N",
            "g": "NC_000009.11:g.21216616T>N",
            "n": "NM_002173.2:n.695A>N",
        },
        # ins, del, delins, dup:
        {
            "c": "NM_002173.2:c.*115_*117insNN",
            "g": "NC_000009.11:g.21216618_21216620insNN",
            "n": "NM_002173.2:n.691_693insNN",
        },
        {
            "c": "NM_002173.2:c.*114_*117delTTCA",
            "g": "NC_000009.11:g.21216618_21216621delGAAA",
            "n": "NM_002173.2:n.690_693delTTCA",
        },
        {
            "c": "NM_002173.2:c.*115_*117delTCAinsNN",
            "g": "NC_000009.11:g.21216618_21216620delGAAinsNN",
            "n": "NM_002173.2:n.691_693delTCAinsNN",
        },
        {
            "c": "NM_002173.2:c.*115_*117delTCAinsNN",
            "g": "NC_000009.11:g.21216618_21216620delGAAinsNN",
            "n": "NM_002173.2:n.691_693delTCAinsNN",
        },
        {
            "c": "NM_002173.2:c.*115_*117dupTCA",
            "g": "NC_000009.11:g.21216618_21216620dupGAA",
            "n": "NM_002173.2:n.691_693dupTCA",
        },
        # genomic variation within an indel discrepancy
        # NM_032790.3  c        108 >
        # NM_032790.3  n        301 > CGGGGAGCCCCCGGGGGCC------CCGCCACCGCCGCCGT
        #                             |||||||||||||||||||------||||||||||||||||
        # NC_000012.11 g  122064755 > CGGGGAGCCCCCGGGGGCCCCGCCACCGCCACCGCCGCCGT
        #                                              **********
        {
            "c": "NM_032790.3:c.125_128delCCCC",
            "g": "NC_000012.11:g.122064772_122064781delCCCCGCCACC",
            "n": "NM_032790.3:n.318_321delCCCC",
        },
    ]

    @classmethod
    def setUpClass(cls):
        def _parse_rec(rec):
            rec["pv"] = {x: cls.hp.parse_hgvs_variant(rec[x]) for x in "cgn"}
            return rec

        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        cls.am = hgvs.assemblymapper.AssemblyMapper(
            cls.hdp, replace_reference=True, assembly_name="GRCh37", alt_aln_method="splign"
        )
        cls.hp = hgvs.parser.Parser()
        cls.tests = [_parse_rec(rec) for rec in cls.test_cases]

    def test_replace_reference_sequence(self):
        """AssemblyMapper: Replace invalid reference sequence"""

        for rec in self.tests:
            for x in "cgn":
                pv = rec["pv"][x]
                if pv.posedit.edit.ref:
                    # replace ref with junk
                    pv.posedit.edit.ref = "NNNNNN"
                self.am._replace_reference(pv)
                self.assertEqual(rec[x], pv.format(conf={"max_ref_length": None}))


@pytest.mark.quick
class Test_AssemblyMapper(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        cls.hp = hgvs.parser.Parser()
        cls.am = hgvs.assemblymapper.AssemblyMapper(
            cls.hdp, assembly_name="GRCh37", alt_aln_method="splign"
        )

    def _test_mapping(self, hgvs_set):
        """given list of variant strings, test all valid combinations of
        g<->n<->c<->p mappings

        """
        parsed_variants = [(hv, self.hp.parse_hgvs_variant(hv)) for hv in hgvs_set]
        hgvs = {v.type: hv for hv, v in parsed_variants}
        pvs = {v.type: v for hv, v in parsed_variants}

        if "g" in pvs and "c" in pvs:
            self.assertEqual(hgvs["g"], str(self.am.c_to_g(pvs["c"])))
            self.assertEqual(hgvs["c"], str(self.am.g_to_c(pvs["g"], pvs["c"].ac)))
        if "g" in pvs and "n" in pvs:
            self.assertEqual(hgvs["g"], str(self.am.n_to_g(pvs["n"])))
            self.assertEqual(hgvs["n"], str(self.am.g_to_n(pvs["g"], pvs["n"].ac)))
        if "c" in pvs and "n" in pvs:
            self.assertEqual(hgvs["n"], str(self.am.c_to_n(pvs["c"])))
            self.assertEqual(hgvs["c"], str(self.am.n_to_c(pvs["n"])))
        if "c" in pvs and "p" in pvs:
            self.assertEqual(hgvs["p"], str(self.am.c_to_p(pvs["c"])))
            self.assertEqual(hgvs["p"], str(self.am.t_to_p(pvs["c"])))

    def test_SNV(self):
        """AssemblyMapper: smoketest with SNVs"""
        hgvs_set = [
            "NC_000007.13:g.36561662C>T",
            "NM_001637.3:c.1582G>A",
            "NM_001637.3:n.1983G>A",
            "NP_001628.1:p.(Gly528Arg)",
        ]
        self._test_mapping(hgvs_set)

    def test_intronic(self):
        """AssemblyMapper: smoketest with intronic SNVs"""
        hgvs_set = [
            "NC_000010.10:g.89711873A>C",
            "NM_000314.4:c.493-2A>C",
            "NM_000314.4:n.1524-2A>C",
            "NP_000305.3:p.?",
        ]
        self._test_mapping(hgvs_set)

    def test_t_to_p(self):
        assert "non-coding" == str(self.am.t_to_p(self.hp.parse("NR_027676.1:n.3980del")))
        assert "NP_000050.2:p.(Lys2597=)" == str(
            self.am.t_to_p(self.hp.parse("NM_000059.3:c.7791A>G"))
        )

    def test_issue_704_set_prevalidation_level(self):
        # Doesn't set prevalidation_level

        for prevalidation_level in PrevalidationLevel:
            am = hgvs.assemblymapper.AssemblyMapper(
                self.hdp, replace_reference=True, assembly_name="GRCh37", alt_aln_method="splign",
                normalize=True, prevalidation_level=prevalidation_level.name,
            )
            vm = am._norm.vm
            self.assertEqual(vm.prevalidation_level, prevalidation_level,
                             "AssemblyMapper Normalizer has 'prevalidation_level' set")


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



import os
from pathlib import Path

import pytest
from support import CACHE

from hgvs.assemblymapper import AssemblyMapper
import hgvs.easy
from hgvs.extras.babelfish import Babelfish
from hgvs.variantmapper import VariantMapper


@pytest.fixture(scope="function")
def vcr_config(request):
    """See https://pytest-vcr.readthedocs.io/en/latest/configuration/"""
    test_file_path = Path(request.node.fspath)
    return {
        "cassette_library_dir": str(test_file_path.with_name("cassettes") / test_file_path.stem),
        "record_mode": os.environ.get("VCR_RECORD_MODE", "new_episodes"),
        "cassette_name": f"{request.node.name}.yaml"
    }


hgvs.easy.hdp = hgvs.dataproviders.uta.connect(
    mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
)


@pytest.fixture(scope="session", autouse=True)
def parser():
    return hgvs.easy.parser


@pytest.fixture(scope="session")
def vm(hdp):
    return VariantMapper(hdp)


@pytest.fixture(scope="session")
def am37(hdp):
    return AssemblyMapper(hdp, assembly_name="GRCh37")


@pytest.fixture(scope="session")
def am38(hdp):
    return AssemblyMapper(hdp, assembly_name="GRCh38")


@pytest.fixture(scope="session")
def hdp():
    return hgvs.easy.hdp


@pytest.fixture(scope="session")
def babelfish38(hdp):
    return Babelfish(hdp, assembly_name="GRCh38")


def pytest_report_header(config):
    env_vars = ["UTA_DB_URL", "HGVS_SEQREPO_URL", "HGVS_CACHE_MODE"]
    rv = [f"{ev}: {os.environ.get(ev)}" for ev in sorted(env_vars)]
    rv += [f"hgvs.easy.hdp={hgvs.easy.hdp.url}"]
    rv += [f"{hgvs.easy.hdp.seqfetcher.source=}"]
    return "\n".join(rv)


@pytest.fixture(scope="class")
def kitchen_sink_setup(request, hdp, parser, am37, am38):
    """
    Adds common fixtures to an instance passed via request.cls.

    For example:

    @pytst.mark.usefixtures("kitchen_sink_setup")
    class MyTest(unittest.TestCase):
        def test_foo(self):
             self.parser.parse(...)

    The class must be subclass of unittest.TestCase

    See https://docs.pytest.org/en/5.3.5/unittest.html#mixing-pytest-fixtures-into-unittest-testcase-subclasses-using-marks

    """

    request.cls.hdp = hdp
    request.cls.parser = parser
    request.cls.am37 = am37
    request.cls.am38 = am38




from abc import ABC, abstractmethod

from hgvs.pretty.console.constants import ENDC, TPURPLE, TYELLOW
from hgvs.pretty.models import PrettyConfig


def colorize_hgvs(hgvs_str: str) -> str:
    """Takes a string representation of a hgvs Sequence Variant and renders it with console colors."""

    spl = hgvs_str.split(":")
    var_str = TPURPLE + spl[0] + ENDC
    var_str += ":"

    sec = spl[1].split(".")
    var_str += TYELLOW + sec[0] + ENDC
    var_str += "."
    var_str += sec[1]

    return var_str


class BasicRenderer(ABC):
    """
    BasicRenderer is an abstract base class that provides a template for rendering objects with a specific configuration and orientation.
    Attributes:
        config: Configuration settings for the renderer.
        orientation (int): Orientation setting for the renderer.
    Methods:
        legend():
            Abstract method to generate a legend for the rendered object.
        display():
            Abstract method to display the rendered object.
    """

    def __init__(self, config: PrettyConfig, orientation: int):
        self.config = config
        self.orientation = orientation

    @abstractmethod
    def legend(self):
        pass

    @abstractmethod
    def display(self):
        pass



from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class TxRefDisagreeRenderer(BasicRenderer):
    """
    TxRefDisagreeRenderer is a class that extends BasicRenderer to provide functionality for rendering
    differences between transcript (tx) and reference (ref) genome sequences aka "tx-ref-disagree".
    Methods:
        legend() -> str:
            Returns a string representing the legend for the renderer.
        display(data: VariantData) -> str:
            Displays differences between the transcript and reference genome sequences based on the provided
            VariantData. If there are no differences, an empty string is returned.
            Args:
                data (VariantData): The data containing variant information and position details.
            Returns:
                str: A string representation of the differences, with optional color coding if enabled in the
                configuration.
    """

    def legend(self) -> str:
        return "tx ref dif: "

    def display(self, data: VariantData) -> str:
        """show differences between tx and ref genome, if there are any"""

        if not data.var_c_or_n:
            return ""

        from hgvs.pretty.console.constants import ENDC, COLOR_MAP

        var_str = ""

        for pdata in data.position_details:
            c_offset = pdata.c_offset
            if not pdata.mapped_pos:
                var_str += " "
                continue

            cig = pdata.cigar_ref
            if cig == "=":
                var_str += " "
            elif cig == "N" or c_offset != 0:
                var_str += " "
                continue
            else:
                # an alignment issue, show cigar string
                if self.config.use_color:
                    var_str += COLOR_MAP["tx_ref_disagree"] + cig + ENDC
                else:
                    var_str += cig

        if var_str.isspace():
            return ""

        return var_str



from typing import List, Tuple

import hgvs
from hgvs.assemblymapper import AssemblyMapper
from hgvs.pretty.console.renderer import colorize_hgvs
from hgvs.pretty.datacompiler import DataCompiler
from hgvs.pretty.models import PrettyConfig
from hgvs.pretty.console.chromseqrenderer import ChromSeqRendered
from hgvs.pretty.console.chromseqreverserenderer import ChromReverseSeqRendered
from hgvs.pretty.console.posinfo import ChrPositionInfo
from hgvs.pretty.console.protmappingrenderer import ProtMappingRenderer
from hgvs.pretty.console.protrulerrenderer import ProtRulerRenderer
from hgvs.pretty.console.protseqrenderer import ProtSeqRenderer
from hgvs.pretty.console.ruler import ChrRuler
from hgvs.pretty.console.shuffledvariant import RegionImpacted
from hgvs.pretty.console.txaligrenderer import TxAligRenderer
from hgvs.pretty.console.txmappingrenderer import TxMappingRenderer
from hgvs.pretty.console.txpos import TxRulerRenderer
from hgvs.pretty.console.txrefdisagreerenderer import TxRefDisagreeRenderer
from hgvs.repeats import RepeatAnalyser
from hgvs.sequencevariant import SequenceVariant


class PrettyPrint:
    """A class to handle the pretty printing of HGVS variants with various configurations.

    Attributes:
        config (PrettyConfig): Configuration object for pretty printing.
    Methods:
        __init__(hdp, assembly_mapper, padding_left=20, padding_right=20, use_color=False, show_legend=True, infer_hgvs_c=True, all=False, show_reverse_strand=False, alt_aln_method='splign', reverse_display=True):
            Initializes the PrettyPrint object with the given parameters.
        _get_all_transcripts(var_g) -> List[str]:
            Retrieves all relevant transcripts for a given genomic variant.
        _infer_hgvs_c(var_g: SequenceVariant, tx_ac: str = None) -> SequenceVariant:
            Infers the HGVS coding or non-coding variant from a genomic variant.
        _map_to_chrom(sv: SequenceVariant) -> SequenceVariant:
            Maps a variant to chromosomal coordinates if needed.
        get_hgvs_names(sv: SequenceVariant, tx_ac: str = None) -> Tuple[SequenceVariant, SequenceVariant]:
            Retrieves the HGVS names for a given sequence variant.
        display(sv: SequenceVariant, tx_ac: str = None, display_start: int = None, display_end: int = None) -> str:
            Takes a variant and prints the genomic context around it.
        create_repre(var_g: SequenceVariant, var_c_or_n: SequenceVariant, display_start: int, display_end: int, data_compiler: DataCompiler) -> str:
            Creates a representation of the variant with the given parameters.
    """

    def __init__(
        self,
        hdp: hgvs.dataproviders.interface.Interface,
        assembly_mapper: AssemblyMapper,
        padding_left: int = 20,
        padding_right: int = 20,
        use_color=False,
        show_legend=True,
        infer_hgvs_c=True,
        all=False,
        show_reverse_strand=False,
        alt_aln_method="splign",
        reverse_display=True,
    ):
        """
        :param hdp: HGVS Data Provider Interface-compliant instance
           (see :class:`hgvs.dataproviders.interface.Interface`)

        :param padding: spacing left and right of the variant for display purposes.
        """

        self.config = PrettyConfig(
            hdp=hdp,
            assembly_mapper=assembly_mapper,
            padding_left=padding_left,
            padding_right=padding_right,
            use_color=use_color,
            show_legend=show_legend,
            infer_hgvs_c=infer_hgvs_c,
            all=all,
            show_reverse_strand=show_reverse_strand,
            alt_aln_method=alt_aln_method,
            reverse_display=reverse_display,
        )

    def _get_all_transcripts(self, var_g) -> List[str]:
        transcripts = self.config.assembly_mapper.relevant_transcripts(var_g)

        return transcripts

    def _infer_hgvs_c(
        self, var_g: SequenceVariant, tx_ac: str = None
    ) -> SequenceVariant:
        if not tx_ac:
            transcripts = self._get_all_transcripts(var_g)
            if transcripts:
                tx_ac = transcripts[0]
            else:
                return None

        am = self.config.assembly_mapper

        if tx_ac.startswith("NR_"):
            var_n = am.g_to_n(var_g, tx_ac)
            return var_n
        var_c = am.g_to_c(var_g, tx_ac)
        return var_c

    def _map_to_chrom(self, sv: SequenceVariant) -> SequenceVariant:
        """maps a variant to chromosomal coords, if needed."""
        am = self.config.assembly_mapper

        if sv.type == "c":
            return am.c_to_g(sv)
        elif sv.type == "n":
            return am.n_to_g(sv)
        elif sv.type == "t":
            return am.t_to_g(sv)
        elif sv.type == "r":
            return am.r

    def get_hgvs_names(
        self, sv: SequenceVariant, tx_ac: str = None
    ) -> Tuple[SequenceVariant, SequenceVariant]:
        var_c_or_n = None
        if sv.type == "g":
            var_g = sv
            if tx_ac is not None:
                var_c_or_n = self._infer_hgvs_c(var_g, tx_ac=tx_ac)
        elif sv.type == "c":
            var_g = self._map_to_chrom(sv)
            var_c_or_n = sv
        elif sv.type == "r":
            var_g = self._map_to_chrom(sv)
            var_c_or_n = sv
        elif sv.type == "n":
            # map back to genome
            var_g = self._map_to_chrom(sv)
            var_c_or_n = sv

        return var_g, var_c_or_n

    def display(
        self,
        sv: SequenceVariant,
        tx_ac: str = None,
        display_start: int = None,
        display_end: int = None,
    ) -> str:
        """Takes a variant and prints the genomic context around it in a string representation."""

        var_g, var_c_or_n = self.get_hgvs_names(sv, tx_ac)

        self.data_compiler = DataCompiler(config=self.config)

        if self.config.all:
            # get all overlapping transcripts

            response = ""
            tx_acs = self._get_all_transcripts(var_g)
            for tx_ac in tx_acs:
                var_c_or_n = self._infer_hgvs_c(var_g, tx_ac)
                response += self.create_repre(
                    var_g, var_c_or_n, display_start, display_end, self.data_compiler
                )
                response += "\n---\n"
            return response
        else:
            if not var_c_or_n:
                if self.config.infer_hgvs_c:
                    var_c_or_n = self._infer_hgvs_c(var_g)

            return self.create_repre(
                var_g, var_c_or_n, display_start, display_end, self.data_compiler
            )

    def create_repre(
        self,
        var_g: SequenceVariant,
        var_c_or_n: SequenceVariant,
        display_start: int,
        display_end: int,
        data_compiler: DataCompiler,
    ):
        """
        Creates a string representation of sequence variants with optional colorization and legends.

        The string representation is built by calling various renderers to display the genomic, coding, and protein
        sequences, as well as the alignment of the sequences. The display is formatted with optional colorization.

        Configuration is possible through the PrettyConfig object, which has been passed into the constructor.

        Args:
            var_g (SequenceVariant): Genomic sequence variant.
            var_c_or_n (SequenceVariant): Coding or non-coding sequence variant.
            display_start (int): Start position for display.
            display_end (int): End position for display.
            data_compiler (DataCompiler): Compiler for sequence variant data.
        Returns:
            str: Formatted string representation of the sequence variants.
        """

        data = data_compiler.data(var_g, var_c_or_n, display_start, display_end)

        left_shuffled_var = data.left_shuffled
        right_shuffled_var = data.right_shuffled
        fully_justified_var = data.fully_justified

        if self.config.show_legend:
            head = "hgvs_g    : "
            head_c = "hgvs_c    : "
            head_n = "hgvs_n    : "
            head_p = "hgvs_p    : "
            refa = "ref>alt   : "
        else:
            head = head_c = head_p = refa = ""

        if self.config.use_color:
            var_g_print = colorize_hgvs(str(var_g))
        else:
            var_g_print = str(var_g)

        var_str = head + var_g_print + "\n"
        if data.var_c_or_n:
            if self.config.use_color:
                var_c_print = colorize_hgvs(str(var_c_or_n))
            else:
                var_c_print = str(var_c_or_n)
            if data.var_c_or_n.type == "c":
                var_str += head_c
            elif data.var_c_or_n.type == "n":
                var_str += head_n
            var_str += var_c_print + "\n"

        if data.var_p:
            if self.config.use_color:
                var_p_print = colorize_hgvs(str(data.var_p))
            else:
                var_p_print = str(data.var_p)
            var_str += head_p + var_p_print + "\n"

        renderer_cls = [ChrPositionInfo, ChrRuler, ChromSeqRendered]

        # if we show reverse strand transcripts in forward facing orientation, always
        # show both forward and reverse strand sequences.
        if (
            self.config.show_reverse_strand
            or self.config.reverse_display
            and data.strand < 0
        ):
            renderer_cls.append(ChromReverseSeqRendered)

        renderer_cls.append(TxRefDisagreeRenderer)

        renderers = []
        for cls in renderer_cls:
            r = cls(self.config, data.strand)
            renderers.append(r)

        for renderer in renderers:
            d = ""
            if self.config.show_legend:
                d += renderer.legend()
            str_results = renderer.display(data)

            if str_results:
                var_str += d + str_results + "\n"

        left_shuffled_renderer = RegionImpacted(
            self.config, data.strand, var_g, left_shuffled_var
        )
        left_shuffled_str = left_shuffled_renderer.display(data)

        right_shuffled_renderer = RegionImpacted(
            self.config, data.strand, var_g, right_shuffled_var
        )
        right_shuffled_str = right_shuffled_renderer.display(data)
        if self.config.show_legend:
            shuffled_seq_header = left_shuffled_renderer.legend()
        else:
            shuffled_seq_header = ""

        if left_shuffled_str != right_shuffled_str:
            fully_justified_renderer = RegionImpacted(
                self.config, data.strand, var_g, fully_justified_var
            )
            fully_justified_str = fully_justified_renderer.display(data)
            if self.config.show_all_shuffleable_regions:
                var_str += shuffled_seq_header + left_shuffled_str + "\n"
                var_str += shuffled_seq_header + right_shuffled_str + "\n"
            var_str += shuffled_seq_header + fully_justified_str + "\n"

        else:
            var_str += shuffled_seq_header + left_shuffled_str + "\n"

        renderers_cls = [
            TxAligRenderer,
            TxMappingRenderer,
            TxRulerRenderer,
            ProtSeqRenderer,
            ProtMappingRenderer,
            ProtRulerRenderer,
        ]
        for cls in renderers_cls:
            renderer = cls(self.config, data.strand)

            d = ""
            if self.config.show_legend:
                d += renderer.legend()
            str_results = renderer.display(data)

            if str_results:
                var_str += d + str_results + "\n"

        if left_shuffled_str != right_shuffled_str:
            # TODO: detect repeats?
            fully_justified_ref = fully_justified_var.ref
            fully_justified_alt = fully_justified_var.alt

            ra = RepeatAnalyser(fully_justified_var)
            if ra.is_repeat:
                var_str += refa + str(ra)
            else:
                var_str += refa + f"{fully_justified_ref}>{fully_justified_alt}"

        return var_str



from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class ProtRulerRenderer(BasicRenderer):
    """
    ProtRulerRenderer is a class that extends BasicRenderer to provide
    functionality for rendering protein sequence positions in p (amino acid) coordinates.
    Methods
    -------
    legend() -> str
        Returns a string representing the legend for the protein ruler.
    display(data: VariantData) -> str
        Generates a string that visually represents the positions of the protein
        sequence based on the provided VariantData.
    """

    def legend(self) -> str:
        return "          : "

    def display(self, data: VariantData) -> str:
        """show the position of the protein seq"""
        var_str = ""

        count = -1
        prev_aa = -1
        for pdata in data.position_details:
            count += 1

            if not pdata.mapped_pos:
                var_str += " "
                prev_aa = -1
                continue

            prot_data = pdata.protein_data

            if not prot_data or prot_data.aa_pos < 0:
                var_str += " "
                prev_aa = -1
                continue

            if len(var_str) > count:
                continue

            aa_pos = prot_data.aa_pos

            if aa_pos == prev_aa:
                var_str += " "
                continue

            prev_aa = aa_pos

            if (aa_pos + 1) % 10 == 0:
                var_str += f"{aa_pos + 1} "
                continue

            elif aa_pos == 0:
                var_str += f"{aa_pos + 1} "
                continue

            var_str += " "

        return var_str



# -*- coding: utf-8 -*-
"""Provides classes for dealing with the locations of HGVS variants

This module provides for Representing the location of variants in HGVS nomenclature, including:

  * integers and integer intervals (e.g., NC_012345.6:g.3403243_3403248A>C)
  * CDS positions and intervals (e.g., NM_01234.5:c.56+12_56+14delAC)
  * CDS stop coordinates (e.g., NM_01234.5:c.*13A>C)

Classes:

  * :class:`SimplePosition` -- a simple integer
  * :class:`BaseOffsetPosition` -- a position with datum, base, and offset for c. and r. coordinates
  * :class:`AAPosition` -- an amino acid position (with AA)
  * :class:`Interval` -- an interval of Positions
"""

from functools import total_ordering

import attr
from bioutils.sequences import aa1_to_aa3

import hgvs
from hgvs.enums import Datum, ValidationLevel
from hgvs.exceptions import HGVSInvalidIntervalError, HGVSUnsupportedOperationError


@attr.s(slots=True, repr=False, cmp=False)
@total_ordering
class SimplePosition:
    base = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def __str__(self):
        self.validate()
        s = "?" if self.base is None else str(self.base)
        return "(" + s + ")" if self.uncertain else s

    def format(self, conf):
        return str(self)

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join((a.name + "=" + str(getattr(self, a.name))) for a in self.__attrs_attrs__),
        )

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base is None

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    def validate(self):
        if self.base is not None and self.base < 1:
            return (ValidationLevel.ERROR, "Position base must be >= 1")
        return (ValidationLevel.VALID, None)

    def __sub__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot substract coordinates of different representations"
        return lhs.base - rhs.base

    def __eq__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        return lhs.base == rhs.base

    def __lt__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        return lhs.base < rhs.base


@attr.s(slots=True, repr=False, cmp=False)
@total_ordering
class BaseOffsetPosition:
    """
    Class for dealing with CDS coordinates in transcript variants.

    This class models CDS positions using a `base` coordinate, which is
    measured relative to a specified `datum` (CDS_START or CDS_END), and
    an `offset`, which is 0 for exonic positions and non-zero for intronic
    positions.  **Positions and offsets are 1-based**, with no 0, per the HGVS
    recommendations.  (If you"re using this with UTA, be aware that UTA
    uses interbase coordinates.)

    +----------+------------+-------+---------+------------------------------------------+
    | hgvs     | datum      | base  | offset  | meaning                                  |
    +==========+============+=======+=========+==========================================+
    | r.55     | SEQ_START  |   55  |      0  | RNA position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55     | CDS_START  |   55  |      0  | CDS position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55     | CDS_START  |   55  |      0  | CDS position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55+1   | CDS_START  |   55  |      1  | intronic variant +1 from boundary        |
    +----------+------------+-------+---------+------------------------------------------+
    | c.-55    | CDS_START  |  -55  |      0  | 5' UTR variant, 55 nt upstream of ATG    |
    +----------+------------+-------+---------+------------------------------------------+
    | c.1      | CDS_START  |    1  |      0  | start codon                              |
    +----------+------------+-------+---------+------------------------------------------+
    | c.1234   | CDS_START  | 1234  |      0  | stop codon (assuming CDS length is 1233) |
    +----------+------------+-------+---------+------------------------------------------+
    | c.*1     | CDS_END    |    0  |      1  | STOP + 1                                 |
    +----------+------------+-------+---------+------------------------------------------+
    | c.*55    | CDS_END    |    0  |     55  | 3' UTR variant, 55 nt after STOP         |
    +----------+------------+-------+---------+------------------------------------------+
    """

    base = attr.ib(default=None)
    offset = attr.ib(default=0)
    datum = attr.ib(default=Datum.SEQ_START)
    uncertain = attr.ib(default=False)

    def validate(self):
        if self.base is not None and self.base == 0:
            return (ValidationLevel.ERROR, "BaseOffsetPosition base may not be 0")
        if (
            hgvs.global_config.mapping.strict_bounds
            and self.base is not None
            and self.datum != Datum.CDS_START
            and self.base < 1
        ):
            return (
                ValidationLevel.ERROR,
                "BaseOffsetPosition base must be >=1 for datum = SEQ_START or CDS_END",
            )
        return (ValidationLevel.VALID, None)

    def __str__(self):
        self.validate()
        base_str = (
            "?"
            if self.base is None
            else "*" + str(self.base)
            if self.datum == Datum.CDS_END
            else str(self.base)
        )
        offset_str = (
            "+?" if self.offset is None else "" if self.offset == 0 else "%+d" % self.offset
        )
        pos = base_str + offset_str
        return "(" + pos + ")" if self.uncertain else pos

    def format(self, conf):
        return str(self)

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join((a.name + "=" + str(getattr(self, a.name))) for a in self.__attrs_attrs__),
        )

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base is None or self.offset is None

    @property
    def is_intronic(self):
        """returns True if the variant is intronic (if the offset is None or non-zero)"""
        return self.offset is None or self.offset != 0

    def __sub__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot substract coordinates of different representations"
        if lhs.datum != rhs.datum:
            raise HGVSUnsupportedOperationError(
                "Interval length measured from different datums is ill-defined"
            )
        if lhs.base == rhs.base:
            return lhs.offset - rhs.offset
        if lhs.offset != 0 or rhs.offset != 0:
            raise HGVSUnsupportedOperationError(
                "Interval length with intronic offsets is ill-defined"
            )
        straddles_zero = 1 if (lhs.base > 0 and rhs.base < 0) else 0
        return lhs.base - rhs.base - straddles_zero

    def __eq__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        return lhs.datum == rhs.datum and lhs.base == rhs.base and lhs.offset == rhs.offset

    def __lt__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        if lhs.datum == rhs.datum:
            if lhs.base == rhs.base:
                return lhs.offset < rhs.offset
            else:
                if (rhs.base - lhs.base == 1 and lhs.offset > 0 and rhs.offset < 0) or (
                    lhs.base - rhs.base == 1 and rhs.offset > 0 and lhs.offset < 0
                ):
                    raise HGVSUnsupportedOperationError(
                        "Cannot compare coordinates in the same intron with one based on end of exon and the other based on start of next exon"  # noqa: E501
                    )
                else:
                    return lhs.base < rhs.base
        else:
            if lhs.datum == Datum.SEQ_START or rhs.datum == Datum.SEQ_START:
                raise HGVSUnsupportedOperationError(
                    "Cannot compare coordinates of datum SEQ_START with CDS_START or CDS_END"
                )
            else:
                return lhs.datum < rhs.datum


@attr.s(slots=True, repr=False, cmp=False)
class AAPosition:
    base = attr.ib(default=None)
    aa = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def validate(self):
        if self.base is not None and self.base != "" and self.base < 1:
            return (ValidationLevel.ERROR, "AAPosition location must be >=1")
        if self.aa is not None and len(self.aa) > 1:
            return (ValidationLevel.ERROR, "More than 1 AA associated with position")
        return (ValidationLevel.VALID, None)

    def format(self, conf=None):
        self.validate()

        p_3_letter = hgvs.global_config.formatting.p_3_letter
        p_term_asterisk = hgvs.global_config.formatting.p_term_asterisk
        if conf and "p_3_letter" in conf and conf["p_3_letter"] is not None:
            p_3_letter = conf["p_3_letter"]
        if conf and "p_term_asterisk" in conf and conf["p_term_asterisk"] is not None:
            p_term_asterisk = conf["p_term_asterisk"]

        pos = "?" if self.base is None else str(self.base)
        if p_3_letter:
            aa = "?" if self.aa is None else aa1_to_aa3(self.aa)
            if p_term_asterisk and aa == "Ter":
                aa = "*"
        else:
            aa = "?" if self.aa is None else self.aa
        s = aa + pos
        return "(" + s + ")" if self.uncertain else s

    __str__ = format

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join((a.name + "=" + str(getattr(self, a.name))) for a in self.__attrs_attrs__),
        )

    @property
    def pos(self):
        """return base, for backward compatibility"""
        return self.base

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base is None or self.aa is None

    def __sub__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot substract coordinates of different representations"
        return lhs.base - rhs.base

    def __eq__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        return lhs.base == rhs.base and lhs.aa == rhs.aa

    def __lt__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        return lhs.base < rhs.base

    def __gt__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        return lhs.base > rhs.base

    def __le__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        return lhs.base <= rhs.base

    def __ge__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot compare coordinates of different representations"
        if lhs.uncertain or rhs.uncertain:
            raise HGVSUnsupportedOperationError("Cannot compare coordinates of uncertain positions")
        return lhs.base >= rhs.base


@attr.s(slots=True, repr=False)
class Interval:
    start = attr.ib(default=None)
    end = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def validate(self):
        if self.start:
            (res, msg) = self.start.validate()
            if res != ValidationLevel.VALID:
                return (res, msg)
        if self.end:
            (res, msg) = self.end.validate()
            if res != ValidationLevel.VALID:
                return (res, msg)
        # Check start less than or equal to end
        if not self.start or not self.end:
            return (ValidationLevel.VALID, None)
        try:
            if self.start <= self.end:
                return (ValidationLevel.VALID, None)
            else:
                return (ValidationLevel.ERROR, "base start position must be <= end position")
        except HGVSUnsupportedOperationError as err:
            return (ValidationLevel.WARNING, str(err))

    def format(self, conf=None):
        if self.start is None:
            return ""
        if self.end is None or self.start == self.end:
            return self.start.format(conf)
        iv = self.start.format(conf) + "_" + self.end.format(conf)
        return "(" + iv + ")" if self.uncertain else iv

    __str__ = format

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join((a.name + "=" + str(getattr(self, a.name))) for a in self.__attrs_attrs__),
        )

    def _set_uncertain(self):
        "mark this interval as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    def _length(self):
        return 1 if self.end is None else self.end - self.start + 1

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.start.is_uncertain or self.end.is_uncertain


@attr.s(slots=True, repr=False)
class BaseOffsetInterval(Interval):
    """BaseOffsetInterval isa Interval of BaseOffsetPositions.  The only
    additional functionality over Interval is to ensure that the dutum
    of end and start are compatible.

    """

    def __attrs_post_init__(self):
        # #330: In a post-ter interval like *87_91, the * binds only
        # to the start. This means that the start.datum is CDS_END,
        # but the end.datum is CDS_START (the default).
        if self.start.datum == Datum.CDS_END:
            self.end.datum = Datum.CDS_END
        self.check_datum()

    def check_datum(self):
        # check for valid combinations of start and end datums
        if (self.start.datum, self.end.datum) not in [
            (Datum.SEQ_START, Datum.SEQ_START),
            (Datum.CDS_START, Datum.CDS_START),
            (Datum.CDS_START, Datum.CDS_END),
            (Datum.CDS_END, Datum.CDS_END),
        ]:
            raise HGVSInvalidIntervalError(
                "BaseOffsetInterval start datum and end datum are incompatible"
            )


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



import pytest
from hgvs.pretty.datacompiler import DataCompiler
from hgvs.pretty.models import PrettyConfig, VariantCoords
from hgvs.repeats import RepeatAnalyser, detect_repetitive_block_lengths, get_repeat_str


def to_hgvs_repeat(chrom_ac, fs: VariantCoords, ra: RepeatAnalyser):
    """Experimental method to see how close we can get to HGVS repeat nomenclature.
    Note: In many cases we are not yet creating valid HGVS strings here. However this works at least for some cases.
    (homoploymeric repeats). If a repeat can be shuffled this needs some modification to use the right-aligned
    representation as well as trimming off bases that can be shuffled (they come out as repeats of length 1 at the moment.)
    """
    return f"{chrom_ac}:g.{fs.start + 1}_{fs.end}{ra.alt_str}"


@pytest.mark.skip(reason="would add too much caching burden.")
class TestHGVSExamples:
    """The variants in this class are taken from the HGVS nomenclature page about repeats
    https://hgvs-nomenclature.org/stable/recommendations/DNA/repeated/
    """

    def test_intergenic_repeat(self, parser, hdp):
        # we can't parse HGVS repeat representations yet:
        # NC_000014.8:g.101179660_101179695TG[14]
        # We have to to use an alternative way to describe this variant:
        hgvs_g = "NC_000014.8:g.101179660_101179667del"

        var_g = parser.parse_hgvs_variant(hgvs_g)

        config = PrettyConfig(hdp, None, None)
        dc = DataCompiler(config)

        fs = dc.get_shuffled_variant(var_g, 0)
        ra = RepeatAnalyser(fs)

        assert ra.is_repeat
        assert ra.ref_str == "TG[18]"
        assert ra.alt_str == "TG[14]"
        assert (
            to_hgvs_repeat(var_g.ac, fs, ra)
            == "NC_000014.8:g.101179660_101179695TG[14]"
        )  # a valid HGVS name

    def test_CACNA1_repeat(self, parser, hdp, am38):
        """NM_023035.2:c.6955_6993CAG[26] (or c.6955_6993dup )
        A repeated CAG tri-nucleotide sequence (CAG on reverse strand, CTG on forward strand).
        """
        hgvs_c = "NM_023035.2:c.6955_6993dup"
        var_c = parser.parse_hgvs_variant(hgvs_c)
        var_g = am38.c_to_g(var_c)
        assert str(var_g) == "NC_000019.10:g.13207860_13207898dup"

        config = PrettyConfig(hdp, None, None)
        dc = DataCompiler(config)

        fs = dc.get_shuffled_variant(var_g, 0)

        ra = RepeatAnalyser(fs)
        assert ra.is_repeat
        assert ra.ref_str == "CTG[13]C[1]"
        assert ra.alt_str == "CTG[26]C[1]"
        assert (
            to_hgvs_repeat(var_g.ac, fs, ra)
            == "NC_000019.10:g.13207859_13207898CTG[26]C[1]"
        )  # not valid. It is 5' shuffled and there's a C that can be shuffled which needs to get trimmed.

        # This repeat is shuffle-able. These are alternate representations of each other:
        # CTG[13]C[1]
        # C[1]TGC[13]

        # here we test the alternative shuffled representation:
        ra2 = RepeatAnalyser(fs, reverse=True)
        assert ra2.is_repeat
        assert ra2.ref_str == "C[1]TGC[13]"
        assert ra2.alt_str == "C[1]TGC[26]"
        assert (
            to_hgvs_repeat(var_g.ac, fs, ra2)
            == "NC_000019.10:g.13207859_13207898C[1]TGC[26]"
        )  # not valid, we should trim off the shuffle-able C for that.

    def test_ATXN7_repeat(self, parser, hdp):
        """A repeated AGC tri-nucleotide sequence in the ATXN7 gene on chromosome 3

        Note: In literature, the tri-nucleotide repeat, encoding a poly-Gln repeat on protein level,
        is known as the CAG repeat (CAG is the codon for Gln).

        This repeat is shuffle-able.

        There are three alternative representations that are all equally valid for this repeat:
        GCA[10]G[1]C[1]
        G[1]CAG[10]C[1]
        G[1]C[1]AGC[10]

        At the moment we can create the first and third representation of this repeat, but not yet the middle one. Note: the middle one is the one
        that would match the codons in the transcript, and prob make most sense from a biological perspective.
        """

        hgvs_g = "NC_000003.12:g.63912688_63912689insCAGCAGCAG"
        var_g = parser.parse_hgvs_variant(hgvs_g)

        config = PrettyConfig(hdp, None, None)
        dc = DataCompiler(config)

        fs = dc.get_shuffled_variant(var_g, 0)
        ra = RepeatAnalyser(fs)

        assert ra.is_repeat
        assert ra.ref_str == "GCA[10]G[1]C[1]"
        assert ra.alt_str == "GCA[13]G[1]C[1]"

        # invalid HGVS, 5' aligned, and shows the two shuffle-able bases at the end.
        assert (
            to_hgvs_repeat(var_g.ac, fs, ra)
            == "NC_000003.12:g.63912685_63912716GCA[13]G[1]C[1]"
        )

        ra2 = RepeatAnalyser(fs, reverse=True)
        assert ra2.is_repeat
        assert ra2.ref_str == "G[1]C[1]AGC[10]"
        assert ra2.alt_str == "G[1]C[1]AGC[13]"

        # invalid HGVS, due to the shuffle-able bases, but comes pretty close to what HGVS recommends.
        assert (
            to_hgvs_repeat(var_g.ac, fs, ra2)
            == "NC_000003.12:g.63912685_63912716G[1]C[1]AGC[13]"
        )


@pytest.mark.skip(reason="would add too much caching burden.")
@pytest.mark.parametrize(
    "hgvs_g, is_repeat, repeat_unit, ref_count, alt_count, s",
    [
        ("NC_000019.10:g.45770205del", True, "C", 6, 5, "C[6]>C[5]"),
        ("NC_000019.10:g.45770204_45770205del", True, "C", 6, 4, "C[6]>C[4]"),
        ("NC_000019.10:g.45770205insC", True, "C", 6, 7, "C[6]>C[7]"),
        ("NC_000019.10:g.45770206del", False, None, 0, 0, "A>"),
        ("NC_000007.13:g.36561662dup", True, "C", 2, 3, "C[2]>C[3]"),
    ],
)
class TestHomopolymerRepeats:
    def test_homopolymer(
        self, parser, hdp, hgvs_g, is_repeat, repeat_unit, ref_count, alt_count, s
    ):
        var_g = parser.parse_hgvs_variant(hgvs_g)
        config = PrettyConfig(hdp, None, None)
        dc = DataCompiler(config)
        fs = dc.get_shuffled_variant(var_g, 0)
        ra = RepeatAnalyser(fs)

        assert s == str(ra)
        assert is_repeat == ra.is_repeat

        if repeat_unit is None:
            assert len(ra.repeat_units_alt) == 0
        else:
            assert repeat_unit == ra.repeat_units_alt[0].repeat_unit

        if ref_count == 0:
            assert len(ra.repeat_units_ref) == 0
        else:
            assert ref_count == ra.repeat_units_ref[0].repeat_count

        if alt_count == 0:
            assert len(ra.repeat_units_alt) == 0
        else:
            assert alt_count == ra.repeat_units_alt[0].repeat_count


@pytest.mark.skip(reason="would add too much caching burden.")
@pytest.mark.parametrize(
    "hgvs_g, is_repeat, repeat_unit, ref_count, alt_count, str_repr",
    [
        (
            "NC_000021.8:g.46020668_46020682del",
            False,
            "CTG",
            2,
            2,
            "CTGCTGCGCCCCCAGCTGCTGCGCCCC>CTGCTGCGCCCC",
        ),
        (
            "NC_000012.11:g.33049660_33049680dup",
            False,
            "G",
            5,
            5,
            "GGGGGCTGCCATGGGGCCGGTGGGGGC>GGGGGCTGCCATGGGGCCGGTGGGGGCTGCCATGGGGCCGGTGGGGGC",
        ),
        (
            "NC_000005.10:g.123346517_123346518insATTA",
            True,
            "ATTA",
            2,
            3,
            "ATTA[2]>ATTA[3]",
        ),
        (
            "NC_000005.10:g.123346522_123346525dup",
            True,
            "ATTA",
            2,
            3,
            "ATTA[2]>ATTA[3]",
        ),
        (
            "NC_000019.10:g.45770210_45770212del",
            True,
            "CAG",
            20,
            19,
            "CAG[20]C[1]A[1]>CAG[19]C[1]A[1]",
        ),  # note this one can be shuffled CAG/GCA
        ("NC_000007.14:g.117548628_117548629insTTTT", True, "T", 7, 11, "T[7]>T[11]"),
        ("NC_000009.11:g.35079521_35079523del", True, "TGG", 2, 1, "TGG[2]>TGG[1]"),
        (
            "NC_000001.11:g.6490477_6490484del",
            True,
            "TCTAAGGC",
            2,
            1,
            "TCTAAGGC[2]T[1]C[1]>TCTAAGGC[1]T[1]C[1]",
        ),
    ],
)
class TestRepeats:
    def test_repeats(
        self,
        parser,
        hdp,
        hgvs_g,
        is_repeat,
        repeat_unit,
        ref_count,
        alt_count,
        str_repr,
    ):
        var_g = parser.parse_hgvs_variant(hgvs_g)
        config = PrettyConfig(hdp, None, None)
        dc = DataCompiler(config)
        fs = dc.get_shuffled_variant(var_g, 0)
        ra = RepeatAnalyser(fs)

        assert str_repr == str(ra)

        # note: in some cases is_repeat is False, although repeat_unit contains some data.
        # however in those cases the repeat patter was not considered significant enough to name the whole variant accordingly
        assert is_repeat == ra.is_repeat
        assert repeat_unit == ra.repeat_units_alt[0].repeat_unit
        assert ref_count == ra.repeat_units_ref[0].repeat_count
        assert alt_count == ra.repeat_units_alt[0].repeat_count


@pytest.mark.skip(reason="would add too much caching burden.")
class TestLargeVariant:
    def test_large_variant(self, parser, hdp):
        """Large variants would take too much time to analyze, plus the results are not that useful."""

        hgvs_g = "NC_000001.11:g.9976249_9983617dup"
        var_g = parser.parse_hgvs_variant(hgvs_g)
        config = PrettyConfig(hdp, None, None)
        dc = DataCompiler(config)
        fs = dc.get_shuffled_variant(var_g, 0)
        ra = RepeatAnalyser(fs)
        assert not ra.is_repeat


class TestRepeatMethods:
    @pytest.mark.parametrize(
        "ref, alt, ref_count, ref_repeat_unit, alt_count, alt_repeat_unit, mixed_repeat_ref, mixed_repeat_alt, expected_ref_repeat, expected_alt_repeat",
        [
            ("T", "TTTTTTT", None, None, 7, "T", False, False, "T[1]", "T[7]"),
            ("TTTTTTT", "T", 7, "T", None, None, False, False, "T[7]", "T[1]"),
            ("T", "TT", None, None, 2, "T", False, False, "T[1]", "T[2]"),
            ("TT", "T", 2, "T", None, None, False, False, "T[2]", "T[1]"),
            ("TT", "TT", 2, "T", 2, "T", False, False, "T[2]", "T[2]"),
            ("", "TT", None, None, 2, "T", False, False, None, None),
            ("TT", "", 2, "T", None, None, False, False, None, None),
            ("TT", "TTTTTTT", 2, "T", 7, "T", False, False, "T[2]", "T[7]"),
            ("GA", "GAGA", None, None, 2, "GA", False, False, "GA[1]", "GA[2]"),
            ("TTTT", "TTTTTTT", 4, "T", 7, "T", False, False, "T[4]", "T[7]"),
            ("abc", "abc", None, None, None, None, False, False, None, None),
            ("abcabc", "abcab", 2, "abc", None, None, False, False, "abc[2]", None),
            ("", "", None, None, None, None, False, False, None, None),
            ("GA", "GCGC", None, None, 2, "GC", False, False, None, "GC[2]"),
            ("GCGC", "GA", 2, "GC", None, None, False, False, "GC[2]", None),
            (
                "GAGAAAA",
                "GAGAAAAA",
                2,
                "GA",
                3,
                "A",
                True,
                True,
                "GA[2]A[3]",
                "GA[2]A[4]",
            ),
            (
                "GAGAAAA",
                "GAGAA",
                4,
                "A",
                2,
                "GA",
                True,
                False,
                "GA[2]A[3]",
                "GA[2]A[1]",
            ),
            (
                "GAGAGAABCABCABCGA",
                "ABCABC",
                4,
                "A",
                2,
                "ABC",
                True,
                False,
                "GA[3]ABC[3]G[1]A[1]",
                "ABC[2]",
            ),
            (
                "GAGAGAABCABCABCTGAGA",
                "ABCABC",
                4,
                "A",
                2,
                "ABC",
                True,
                False,
                "GA[3]ABC[3]T[1]GA[2]",
                "ABC[2]",
            ),
            (
                "ABBCABBCABBC",
                "ABBCABBC",
                3,
                "ABBC",
                2,
                "ABBC",
                False,
                False,
                "ABBC[3]",
                "ABBC[2]",
            ),
        ],
    )
    def test_detect_repetitive_block_lengths(
        self,
        ref,
        alt,
        ref_count,
        ref_repeat_unit,
        alt_count,
        alt_repeat_unit,
        mixed_repeat_ref,
        mixed_repeat_alt,
        expected_ref_repeat,
        expected_alt_repeat,
    ):
        repeat_units_ref = detect_repetitive_block_lengths(ref)
        repeat_units_alt = detect_repetitive_block_lengths(alt)

        if ref_count is None:
            assert len(repeat_units_ref) == 0
        if alt_count is None:
            assert len(repeat_units_alt) == 0

        if len(repeat_units_ref) == 1:
            ref_u = repeat_units_ref[0].repeat_unit
            ref_l = repeat_units_ref[0].repeat_count
            assert ref_u == ref_repeat_unit
            assert ref_l == ref_count
            assert not mixed_repeat_ref
        elif len(repeat_units_ref) > 1:
            # more than one repeat found.
            assert mixed_repeat_ref

        if len(repeat_units_alt) == 1:
            alt_u = repeat_units_alt[0].repeat_unit
            alt_l = repeat_units_alt[0].repeat_count
            assert alt_u == alt_repeat_unit
            assert alt_l == alt_count
            assert not mixed_repeat_alt
        elif len(repeat_units_alt) > 1:
            assert mixed_repeat_alt

        # next: rebuild ref and alt strings:
        ref_repeat = get_repeat_str(ref, alt, repeat_units_ref, repeat_units_alt)
        assert ref_repeat == expected_ref_repeat

        alt_repeat = get_repeat_str(alt, ref, repeat_units_alt, repeat_units_ref)
        assert alt_repeat == expected_alt_repeat



# -*- coding: utf-8 -*-

import pickle

protocol = 4


class PersistentDict(dict):
    """Persistent dictionary"""

    def __init__(self, filename, flag="c", *args, **kwds):
        self.filename = filename
        self.flag = flag  # r=readonly, c=create,write,read
        try:
            with open(self.filename, "rb") as f:
                self.update(pickle.load(f))  # noqa: B301,BAN-B301
        except IOError:
            if self.flag == "r":
                raise IOError("Cannot open file " + self.filename)
        dict.__init__(self, *args, **kwds)

    def sync(self):
        if self.flag == "r":
            return
        with open(self.filename, "wb") as f:
            pickle.dump(dict(self), f, protocol)

    def close(self):
        self.sync()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Mapping positions between pairs of sequence alignments

The AlignmentMapper class is at the heart of mapping between aligned sequences.

"""

# Implementation note re: "no-zero correction": HGVS doesn't have a
# 0. Counting is -3, -2, -1, 1, 2, 3 :-/ Coordinate calculations must
# take this discontinuity in c. positions into account.  The
# implementation of imaginary transcript positions creates a second
# discontinuity. (By analogy with c., n.0 is declared to not exist.)
# The strategy used in this code is to use internal c0 and n0
# coordinates, which include 0, for coordinate calculations and to
# translate these to c. and n. positions as needed.
#
#                imag.                                                 imag.
#              upstream     5' UTR           CDS          3' UTR      downstr
#                                     |>
#            - - - - - - ———————————— ||||||||||||||||| ——————————— - - - - - -
#                           a     b     C     D     E     f     g     h     i
#    c.        -4    -3    -2    -1  !  1     2     3  ! *1    *2    *3    *4
#    c0        -4    -3    -2    -1     0     1     2     3     4     5     6
#    n0        -2    -1     0     1     2     3     4     5     6     7     8
#    n.        -2    -1  !  1     2     3     4     5     6     7     8     9
#    g.   ... 123   124   125   126   127   128   129   130   131   132   133 ...
#


from bioutils.coordinates import strand_int_to_pm

import hgvs.location
from hgvs import global_config
from hgvs.enums import Datum
from hgvs.exceptions import (
    HGVSDataNotAvailableError,
    HGVSError,
    HGVSInvalidIntervalError,
    HGVSUsageError,
)
from hgvs.utils import build_tx_cigar
from hgvs.utils.cigarmapper import CIGARMapper


def _zbc_to_hgvs(i):
    """Convert zero-based coordinate to hgvs (1 based, missing zero)"""
    if i >= 0:
        i += 1
    return i


def _hgvs_to_zbc(i):
    """Convert hgvs (1 based, missing zero)"""
    if i >= 1:
        i -= 1
    return i


class AlignmentMapper:
    """Maps hgvs location objects between genomic (g), non-coding (n) and
    cds (c) coordinates according to a CIGAR string.

    :param hdp: HGVS Data Provider Interface-compliant instance (see :class:`hgvs.dataproviders.interface.Interface`)
    :param str tx_ac: string representing transcript accession (e.g., NM_000551.2)
    :param str alt_ac: string representing the reference sequence accession (e.g., NC_000019.10)
    :param str alt_aln_method: string representing the alignment method; valid values depend on data source

    """

    __slots__ = (
        "tx_ac",
        "alt_ac",
        "alt_aln_method",
        "strand",
        "gc_offset",
        "cds_start_i",
        "cds_end_i",
        "tgt_len",
        "cigarmapper",
        "ref_pos",
        "tgt_pos",
        "cigar_op",
    )

    def __init__(self, hdp, tx_ac, alt_ac, alt_aln_method):
        self.tx_ac = tx_ac
        self.alt_ac = alt_ac
        self.alt_aln_method = alt_aln_method

        if self.alt_aln_method != "transcript":
            tx_info = hdp.get_tx_info(self.tx_ac, self.alt_ac, self.alt_aln_method)
            if tx_info is None:
                raise HGVSDataNotAvailableError(
                    "AlignmentMapper(tx_ac={self.tx_ac}, "
                    "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method}): "
                    "No transcript info".format(self=self)
                )

            tx_exons = hdp.get_tx_exons(self.tx_ac, self.alt_ac, self.alt_aln_method)
            if tx_exons is None:
                raise HGVSDataNotAvailableError(
                    "AlignmentMapper(tx_ac={self.tx_ac}, "
                    "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method}): "
                    "No transcript exons".format(self=self)
                )

            # hgvs-386: An assumption when building the cigar string
            # is that exons are adjacent. Assert that here.
            sorted_tx_exons = sorted(tx_exons, key=lambda e: e["ord"])
            for i in range(1, len(sorted_tx_exons)):
                if sorted_tx_exons[i - 1]["tx_end_i"] != sorted_tx_exons[i]["tx_start_i"]:
                    raise HGVSDataNotAvailableError(
                        "AlignmentMapper(tx_ac={self.tx_ac}, "
                        "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method}): "
                        "Exons {a} and {b} are not adjacent".format(self=self, a=i, b=i + 1)
                    )

            self.strand = tx_exons[0]["alt_strand"]
            self.gc_offset = tx_exons[0]["alt_start_i"]
            self.cds_start_i = tx_info["cds_start_i"]
            self.cds_end_i = tx_info["cds_end_i"]

            cigar = build_tx_cigar(tx_exons, self.strand)
            self.cigarmapper = CIGARMapper(cigar)
            self.tgt_len = self.cigarmapper.tgt_len

        else:
            # this covers the identity cases n <-> c
            tx_identity_info = hdp.get_tx_identity_info(self.tx_ac)
            if tx_identity_info is None:
                raise HGVSDataNotAvailableError(
                    "AlignmentMapper(tx_ac={self.tx_ac}, "
                    "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method}): "
                    "No transcript info".format(self=self)
                )
            self.cds_start_i = tx_identity_info["cds_start_i"]
            self.cds_end_i = tx_identity_info["cds_end_i"]
            self.tgt_len = sum(tx_identity_info["lengths"])
            self.cigarmapper = None

        assert not (
            (self.cds_start_i is None) ^ (self.cds_end_i is None)
        ), "CDS start and end must both be defined or neither defined"

    def __str__(self):
        return (
            "{self.__class__.__name__}: {self.tx_ac} ~ {self.alt_ac} ~ {self.alt_aln_method}; "
            "{strand_pm} strand; offset={self.gc_offset}".format(
                self=self, strand_pm=strand_int_to_pm(self.strand)
            )
        )

    def g_to_n(self, g_interval, strict_bounds=None):
        """convert a genomic (g.) interval to a transcript cDNA (n.) interval"""

        if strict_bounds is None:
            strict_bounds = global_config.mapping.strict_bounds

        grs, gre = (
            g_interval.start.base - 1 - self.gc_offset,
            g_interval.end.base - 1 - self.gc_offset,
        )
        # frs, fre = (f)orward (r)na (s)tart & (e)nd; forward w.r.t. genome
        frs, frs_offset, frs_cigar = self.cigarmapper.map_ref_to_tgt(
            pos=grs, end="start", strict_bounds=strict_bounds
        )
        fre, fre_offset, fre_cigar = self.cigarmapper.map_ref_to_tgt(
            pos=gre, end="end", strict_bounds=strict_bounds
        )

        if self.strand == -1:
            frs, fre = self.tgt_len - 1 - fre, self.tgt_len - 1 - frs
            frs_offset, fre_offset = -fre_offset, -frs_offset

        # The returned interval would be uncertain when locating at alignment gaps
        return hgvs.location.BaseOffsetInterval(
            start=hgvs.location.BaseOffsetPosition(
                base=_zbc_to_hgvs(frs), offset=frs_offset, datum=Datum.SEQ_START
            ),
            end=hgvs.location.BaseOffsetPosition(
                base=_zbc_to_hgvs(fre), offset=fre_offset, datum=Datum.SEQ_START
            ),
            uncertain=frs_cigar in "DI" or fre_cigar in "DI",
        )

    def n_to_g(self, n_interval, strict_bounds=None):
        """convert a transcript (n.) interval to a genomic (g.) interval"""

        if strict_bounds is None:
            strict_bounds = global_config.mapping.strict_bounds

        frs = _hgvs_to_zbc(n_interval.start.base)
        start_offset = n_interval.start.offset
        fre = _hgvs_to_zbc(n_interval.end.base)
        end_offset = n_interval.end.offset

        if self.strand == -1:
            fre, frs = self.tgt_len - 1 - frs, self.tgt_len - 1 - fre
            start_offset, end_offset = -end_offset, -start_offset

        # returns the genomic range start (grs) and end (gre)
        grs, _, grs_cigar = self.cigarmapper.map_tgt_to_ref(
            pos=frs, end="start", strict_bounds=strict_bounds
        )
        gre, _, gre_cigar = self.cigarmapper.map_tgt_to_ref(
            pos=fre, end="end", strict_bounds=strict_bounds
        )
        grs, gre = grs + self.gc_offset + 1, gre + self.gc_offset + 1
        gs, ge = grs + start_offset, gre + end_offset

        # The returned interval would be uncertain when locating at alignment gaps
        return hgvs.location.Interval(
            start=hgvs.location.SimplePosition(gs, uncertain=n_interval.start.uncertain),
            end=hgvs.location.SimplePosition(ge, uncertain=n_interval.end.uncertain),
            uncertain=grs_cigar in "DI" or gre_cigar in "DI",
        )

    def n_to_c(self, n_interval, strict_bounds=None):
        """convert a transcript cDNA (n.) interval to a transcript CDS (c.) interval"""

        if strict_bounds is None:
            strict_bounds = global_config.mapping.strict_bounds

        if (
            self.cds_start_i is None
        ):  # cds_start_i defined iff cds_end_i defined; see assertion above
            raise HGVSUsageError(
                "CDS is undefined for {self.tx_ac}; cannot map to c. coordinate (non-coding transcript?)".format(
                    self=self
                )
            )

        if strict_bounds and (n_interval.start.base <= 0 or n_interval.end.base > self.tgt_len):
            raise HGVSInvalidIntervalError(
                "The given coordinate is outside the bounds of the reference sequence."
            )

        def pos_n_to_c(pos):
            if pos.base <= self.cds_start_i:
                c = pos.base - self.cds_start_i - (1 if pos.base > 0 else 0)
                c_datum = Datum.CDS_START
            elif pos.base > self.cds_start_i and pos.base <= self.cds_end_i:
                c = pos.base - self.cds_start_i
                c_datum = Datum.CDS_START
            else:
                c = pos.base - self.cds_end_i
                c_datum = Datum.CDS_END
            return hgvs.location.BaseOffsetPosition(base=c, offset=pos.offset, datum=c_datum)

        c_interval = hgvs.location.BaseOffsetInterval(
            start=pos_n_to_c(n_interval.start),
            end=pos_n_to_c(n_interval.end),
            uncertain=n_interval.uncertain,
        )
        return c_interval

    def c_to_n(self, c_interval, strict_bounds=None):
        """convert a transcript CDS (c.) interval to a transcript cDNA (n.) interval"""

        if strict_bounds is None:
            strict_bounds = global_config.mapping.strict_bounds

        if self.cds_start_i is None:
            raise HGVSUsageError(
                "CDS is undefined for {self.tx_ac}; this accession appears to be for a non-coding transcript".format(
                    self=self
                )
            )

        def pos_c_to_n(pos):
            if pos.datum == Datum.CDS_START:
                n = pos.base + self.cds_start_i
                if pos.base < 0:  # correct for lack of c.0 coordinate
                    n += 1
            elif pos.datum == Datum.CDS_END:
                n = pos.base + self.cds_end_i
            if n <= 0:  # correct for lack of n.0 coordinate
                n -= 1
            if n <= 0 or n > self.tgt_len:
                if strict_bounds:
                    raise HGVSInvalidIntervalError(f"c.{pos} coordinate is out of bounds")
            return hgvs.location.BaseOffsetPosition(
                base=n, offset=pos.offset, datum=Datum.SEQ_START
            )

        n_interval = hgvs.location.BaseOffsetInterval(
            start=pos_c_to_n(c_interval.start),
            end=pos_c_to_n(c_interval.end),
            uncertain=c_interval.uncertain,
        )
        return n_interval

    def g_to_c(self, g_interval, strict_bounds=None):
        """convert a genomic (g.) interval to a transcript CDS (c.) interval"""
        return self.n_to_c(self.g_to_n(g_interval), strict_bounds=strict_bounds)

    def c_to_g(self, c_interval, strict_bounds=None):
        """convert a transcript CDS (c.) interval to a genomic (g.) interval"""
        return self.n_to_g(self.c_to_n(c_interval), strict_bounds=strict_bounds)

    @property
    def is_coding_transcript(self):
        if (self.cds_start_i is not None) ^ (self.cds_end_i is not None):
            raise HGVSError(
                "{self.tx_ac}: CDS start_i and end_i"
                " must be both defined or both undefined".format(self=self)
            )
        return self.cds_start_i is not None

    def g_interval_is_inbounds(self, ival):
        grs = ival.start.base - 1 - self.gc_offset
        gre = ival.end.base - 1 - self.gc_offset
        return grs >= 0 and gre <= self.cigarmapper.ref_len


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class TxMappingRenderer(BasicRenderer):
    """
    TxMappingRenderer is a class that extends BasicRenderer to provide
    functionality for rendering transcript position information by printing a | eveery 10 bases and a . every 5 bases.

    Methods
    -------
    legend():
        Returns a string representing the legend for the transcript position.
    display(data: VariantData) -> str:
        Generates a string representation of the transcript sequence positions
        based on the provided VariantData. The output string uses specific
        characters to denote different positions and offsets within the
        transcript sequence.
    """

    def legend(self):
        return "tx pos    : "

    def display(self, data: VariantData) -> str:
        """show the position of the transcript seq"""

        var_str = ""

        count = -1
        prev_c_pos = ""
        for pdata in data.position_details:
            count += 1
            if not pdata.mapped_pos:
                var_str += " "
                prev_c_pos = ""
                continue

            c_pos = pdata.c_pos
            if c_pos is None and pdata.n_pos:
                c_pos = pdata.n_pos

            if c_pos is None:
                var_str += " "
                prev_c_pos = c_pos
                continue

            if pdata.cigar_ref == "N":
                var_str += " "
                prev_c_pos = c_pos
                continue

            if len(var_str) > count:
                prev_c_pos = c_pos
                continue

            if (c_pos) % 10 == 0:
                var_str += "|"
                prev_c_pos = c_pos
                continue

            elif (c_pos) % 5 == 0:
                var_str += "."
                prev_c_pos = c_pos
                continue

            elif (
                prev_c_pos
                and prev_c_pos == c_pos
                and pdata.c_offset > 0
                and (pdata.c_offset % 10) == 0
            ):
                var_str += "^"
                prev_c_pos = c_pos
                continue
            elif (
                prev_c_pos
                and prev_c_pos == c_pos
                and pdata.c_offset > 0
                and (pdata.c_offset % 5) == 0
            ):
                var_str += "."
                prev_c_pos = c_pos
                continue

            elif c_pos == 1:
                var_str += "|"

            var_str += " "

        return var_str



import pytest

testdata = [
    ('NM_000203.5:c.1957_*3del', 'NP_000194.2:p.(Pro653CysfsTer?)'),
    ('NM_004985.5:c.564_*2del', 'NP_004976.2:p.(Met188IlefsTer8)'),
    ('NM_004333.6:c.2298_*4del', 'NP_004324.2:p.(His766GlnfsTer26)'),
    ('NM_153223.4:c.2955_*2del', 'NP_694955.2:p.(Ser985ArgfsTer15)'),
]

@pytest.mark.parametrize("var_c_str,var_p_str", testdata)
def test_c_to_p(var_c_str, var_p_str, parser, am37):
    var_c = parser.parse(var_c_str)
    var_p = am37.c_to_p(var_c)
    assert str(var_p) == var_p_str
